from __future__ import annotations

import warnings
from typing import Sequence, Optional, Tuple

import argue
import numpy as np
from matplotlib import pyplot as plt
from scipy.interpolate import interp1d

from .roi import HighContrastDiskROI


class MTF:
    """This class will calculate relative MTF"""

    def __init__(
        self,
        lp_spacings: Sequence[float],
        lp_maximums: Sequence[float],
        lp_minimums: Sequence[float],
    ):
        """
        Parameters
        ----------
        lp_spacings : sequence of floats
            These are the physical spacings per unit distance. E.g. 0.1 line pairs/mm.
        lp_maximums : sequence of floats
            These are the maximum values of the sample ROIs.
        lp_minimums : sequence of floats
            These are the minimum values of the sample ROIs.
        """
        self.spacings = lp_spacings
        self.maximums = lp_maximums
        self.minimums = lp_minimums
        self.mtfs = {}
        self.norm_mtfs = {}
        for (spacing, max, min) in zip(lp_spacings, lp_maximums, lp_minimums):
            self.mtfs[spacing] = (max - min) / (max + min)
        # sort according to spacings
        self.mtfs = {k: v for k, v in sorted(self.mtfs.items(), key=lambda x: x[0])}
        for key, value in self.mtfs.items():
            self.norm_mtfs[key] = (
                value / self.mtfs[lp_spacings[0]]
            )  # normalize to first region

        # check that the MTF drops monotonically by measuring the deltas between MTFs
        # if the delta is increasing it means the MTF rose on a subsequent value
        max_delta = np.max(np.diff(list(self.norm_mtfs.values())))
        if max_delta > 0:
            warnings.warn(
                "The MTF does not drop monotonically; be sure the ROIs are correctly aligned."
            )

    @argue.bounds(x=(0, 100))
    def relative_resolution(self, x: float = 50) -> float:
        """Return the line pair value at the given rMTF resolution value.

        Parameters
        ----------
        x : float
            The percentage of the rMTF to determine the line pair value. Must be between 0 and 100.
        """
        f = interp1d(
            list(self.norm_mtfs.values()),
            list(self.norm_mtfs.keys()),
            fill_value="extrapolate",
        )
        mtf = f(x / 100)
        if mtf > max(self.spacings):
            warnings.warn(
                f"MTF resolution wasn't calculated for {x}% that was asked for. The value returned is an extrapolation. Use a higher % MTF to get a non-interpolated value."
            )
        return float(mtf)

    @classmethod
    def from_high_contrast_diskset(
        cls, spacings: Sequence[float], diskset: Sequence[HighContrastDiskROI]
    ) -> MTF:
        """Construct the MTF using high contrast disks from the ROI module."""
        maximums = [roi.max for roi in diskset]
        minimums = [roi.min for roi in diskset]
        return cls(spacings, maximums, minimums)

    def plot(
        self,
        axis: Optional[plt.Axes] = None,
        grid: bool = True,
        x_label: str = "Line pairs / mm",
        y_label: str = "Relative MTF",
        title: str = "RMTF",
        margins: float = 0.05,
        marker: str = "o",
        label: str = "rMTF",
    ) -> Tuple:
        """Plot the Relative MTF.

        Parameters
        ----------
        axis : None, matplotlib.Axes
            The axis to plot the MTF on. If None, will create a new figure.
        """
        if axis is None:
            fig, axis = plt.subplots()
        points = axis.plot(
            list(self.norm_mtfs.keys()),
            list(self.norm_mtfs.values()),
            marker=marker,
            label=label,
        )
        axis.margins(margins)
        axis.grid(grid)
        axis.set_xlabel(x_label)
        axis.set_ylabel(y_label)
        axis.set_title(title)
        return points



import sys

from pylinac.version import __version__

# check python version
if sys.version_info[0] < 3 or sys.version_info[1] < 7:
    raise ValueError(
        "Pylinac is only supported on Python 3.7+. Please update your environment."
    )

# import shortcuts
# core first
from .core import decorators, geometry, image, io, mask, profile, roi, utilities

# alphabetized modules
from .acr import ACRCT, ACRMRILarge
from .ct import CatPhan504, CatPhan600, CatPhan503, CatPhan604
from .quart import QuartDVT
from .core.utilities import clear_data_files, assign2machine
from .field_analysis import (
    FieldAnalysis,
    DeviceFieldAnalysis,
    Protocol,
    Device,
    Edge,
    Interpolation,
    Normalization,
    Centering,
)
from .planar_imaging import (
    LeedsTOR,
    StandardImagingQC3,
    LasVegas,
    DoselabMC2kV,
    DoselabMC2MV,
    StandardImagingQCkV,
    PTWEPIDQC,
    SNCMV,
    SNCkV,
    StandardImagingFC2,
    IMTLRad,
    SNCFSQA,
)
from .log_analyzer import load_log, Dynalog, TrajectoryLog, MachineLogs
from .picketfence import PicketFence  # must be after log analyzer
from .starshot import Starshot
from .vmat import DRMLC, DRGS
from .winston_lutz import WinstonLutz
from .calibration import tg51, trs398



from .simulators import AS1200Image, AS500Image, AS1000Image
from .layers import (
    PerfectBBLayer,
    PerfectConeLayer,
    PerfectFieldLayer,
    GaussianFilterLayer,
    FilterFreeFieldLayer,
    FilteredFieldLayer,
    ConstantLayer,
    FilterFreeConeLayer,
    RandomNoiseLayer,
)
from .utils import generate_picketfence, generate_winstonlutz, generate_winstonlutz_cone



import enum
from typing import Union, Tuple, Optional

import matplotlib.pyplot as plt
import numpy as np
from cached_property import cached_property
from matplotlib.patches import Circle as mpl_Circle
from skimage.measure._regionprops import _RegionProperties

from .decorators import lru_cache
from .geometry import Circle, Point, Rectangle
from .image import ArrayImage


def bbox_center(region: _RegionProperties) -> Point:
    """Return the center of the bounding box of an scikit-image region.

    Parameters
    ----------
    region
        A scikit-image region as calculated by skimage.measure.regionprops().

    Returns
    -------
    point : :class:`~pylinac.core.geometry.Point`
    """
    bbox = region.bbox
    y = abs(bbox[0] - bbox[2]) / 2 + min(bbox[0], bbox[2])
    x = abs(bbox[1] - bbox[3]) / 2 + min(bbox[1], bbox[3])
    return Point(x, y)


class Contrast(enum.Enum):
    """Contrast calculation technique. See :ref:`visibility`"""

    MICHELSON = "Michelson"  #:
    WEBER = "Weber"  #:
    RATIO = "Ratio"  #:


class DiskROI(Circle):
    """An class representing a disk-shaped Region of Interest."""

    def __init__(
        self,
        array: np.ndarray,
        angle: float,
        roi_radius: float,
        dist_from_center: float,
        phantom_center: Union[Tuple, Point],
    ):
        """
        Parameters
        ----------
        array : ndarray
            The 2D array representing the image the disk is on.
        angle : int, float
            The angle of the ROI in degrees from the phantom center.
        roi_radius : int, float
            The radius of the ROI from the center of the phantom.
        dist_from_center : int, float
            The distance of the ROI from the phantom center.
        phantom_center : tuple
            The location of the phantom center.
        """
        center = self._get_shifted_center(angle, dist_from_center, phantom_center)
        super().__init__(center_point=center, radius=roi_radius)
        self._array = array

    @staticmethod
    def _get_shifted_center(
        angle: float,
        dist_from_center: float,
        phantom_center: Point,
    ) -> Point:
        """The center of the ROI; corrects for phantom dislocation and roll."""
        y_shift = np.sin(np.deg2rad(angle)) * dist_from_center
        x_shift = np.cos(np.deg2rad(angle)) * dist_from_center
        return Point(phantom_center.x + x_shift, phantom_center.y + y_shift)

    @cached_property
    def pixel_values(self) -> np.ndarray:
        masked_img = self.circle_mask()
        return self._array[~np.isnan(masked_img)]

    @cached_property
    def pixel_value(self) -> float:
        """The median pixel value of the ROI."""
        masked_img = self.circle_mask()
        return float(np.nanmedian(masked_img))

    @cached_property
    def std(self) -> float:
        """The standard deviation of the pixel values."""
        masked_img = self.circle_mask()
        return float(np.nanstd(masked_img))

    @lru_cache()
    def circle_mask(self) -> np.ndarray:
        """Return a mask of the image, only showing the circular ROI."""
        # http://scikit-image.org/docs/dev/auto_examples/plot_camera_numpy.html
        # TODO: Replace with scikit-image draw function
        masked_array = np.copy(self._array).astype(float)
        l_x, l_y = self._array.shape[0], self._array.shape[1]
        X, Y = np.ogrid[:l_x, :l_y]
        outer_disk_mask = (X - self.center.y) ** 2 + (
            Y - self.center.x
        ) ** 2 > self.radius**2
        masked_array[outer_disk_mask] = np.NaN
        return masked_array

    def plot2axes(
        self, axes: Optional[plt.Axes] = None, edgecolor: str = "black", fill: bool = False
    ) -> None:
        """Plot the Circle on the axes.

        Parameters
        ----------
        axes : matplotlib.axes.Axes
            An MPL axes to plot to.
        edgecolor : str
            The color of the circle.
        fill : bool
            Whether to fill the circle with color or leave hollow.
        """
        if axes is None:
            fig, axes = plt.subplots()
            axes.imshow(self._array)
        axes.add_patch(
            mpl_Circle(
                (self.center.x, self.center.y),
                edgecolor=edgecolor,
                radius=self.radius,
                fill=fill,
            )
        )


class LowContrastDiskROI(DiskROI):
    """A class for analyzing the low-contrast disks."""

    contrast_threshold: Optional[float]
    cnr_threshold: Optional[float]
    contrast_reference: Optional[float]

    def __init__(
        self,
        array: Union[np.ndarray, ArrayImage],
        angle: float,
        roi_radius: float,
        dist_from_center: float,
        phantom_center: Union[tuple, Point],
        contrast_threshold: Optional[float] = None,
        contrast_reference: Optional[float] = None,
        cnr_threshold: Optional[float] = None,
        contrast_method: Contrast = Contrast.MICHELSON,
        visibility_threshold: Optional[float] = 0.1,
    ):
        """
        Parameters
        ----------
        contrast_threshold : float, int
            The threshold for considering a bubble to be "seen".
        """
        super().__init__(array, angle, roi_radius, dist_from_center, phantom_center)
        self.contrast_threshold = contrast_threshold
        self.cnr_threshold = cnr_threshold
        self.contrast_reference = contrast_reference
        self.contrast_method = contrast_method
        self.visibility_threshold = visibility_threshold

    @property
    def signal_to_noise(self) -> float:
        """The signal to noise ratio."""
        return self.pixel_value / self.std

    @property
    def contrast_to_noise(self) -> float:
        """The contrast to noise ratio of the ROI"""
        return self.contrast / self.std

    @property
    def contrast(self) -> float:
        """The contrast of the bubble. Uses the contrast method passed in the constructor. See https://en.wikipedia.org/wiki/Contrast_(vision)."""
        if self.contrast_method == Contrast.MICHELSON:
            return abs(
                (self.pixel_value - self.contrast_reference)
                / (self.pixel_value + self.contrast_reference)
            )
        elif self.contrast_method == Contrast.WEBER:
            return (
                abs(self.pixel_value - self.contrast_reference)
                / self.contrast_reference
            )
        elif self.contrast_method == Contrast.RATIO:
            return self.pixel_value / self.contrast_reference

    @property
    def cnr_constant(self) -> float:
        """The contrast-to-noise value times the bubble diameter."""
        DeprecationWarning(
            "The 'cnr_constant' property will be deprecated in a future release. Use .visibility instead."
        )
        return self.contrast_to_noise * self.diameter

    @property
    def visibility(self) -> float:
        """The visual perception of CNR. Uses the model from A Rose: https://www.osapublishing.org/josa/abstract.cfm?uri=josa-38-2-196.
        See also here: https://howradiologyworks.com/x-ray-cnr/.
        Finally, a review paper here: http://xrm.phys.northwestern.edu/research/pdf_papers/1999/burgess_josaa_1999.pdf
        Importantly, the Rose model is not applicable for high-contrast use cases."""
        return self.contrast * np.sqrt(self.radius**2 * np.pi) / self.std

    @property
    def contrast_constant(self) -> float:
        """The contrast value times the bubble diameter."""
        DeprecationWarning(
            "The 'contrast_constant' property will be deprecated in a future release. Use .visibility instead."
        )
        return self.contrast * self.diameter

    @property
    def passed(self) -> bool:
        """Whether the disk ROI contrast passed."""
        return self.contrast > self.contrast_threshold

    @property
    def passed_visibility(self) -> bool:
        """Whether the disk ROI's visibility passed."""
        return self.visibility > self.visibility_threshold

    @property
    def passed_contrast_constant(self) -> bool:
        """Boolean specifying if ROI pixel value was within tolerance of the nominal value."""
        return self.contrast_constant > self.contrast_threshold

    @property
    def passed_cnr_constant(self) -> bool:
        """Boolean specifying if ROI pixel value was within tolerance of the nominal value."""
        return self.cnr_constant > self.cnr_threshold

    @property
    def plot_color(self) -> str:
        """Return one of two colors depending on if ROI passed."""
        return "green" if self.passed_visibility else "red"

    @property
    def plot_color_constant(self) -> str:
        """Return one of two colors depending on if ROI passed."""
        return "green" if self.passed_contrast_constant else "red"

    @property
    def plot_color_cnr(self) -> str:
        """Return one of two colors depending on if ROI passed."""
        return "green" if self.passed_cnr_constant else "red"


class HighContrastDiskROI(DiskROI):
    """A class for analyzing the high-contrast disks."""

    contrast_threshold: Optional[float]

    def __init__(
        self,
        array: np.ndarray,
        angle: float,
        roi_radius: float,
        dist_from_center: float,
        phantom_center: Union[tuple, Point],
        contrast_threshold: float,
    ):
        """
        Parameters
        ----------
        contrast_threshold : float, int
            The threshold for considering a bubble to be "seen".
        """
        super().__init__(array, angle, roi_radius, dist_from_center, phantom_center)
        self.contrast_threshold = contrast_threshold

    def __repr__(self):
        return f"High-Contrast Disk; max pixel: {self.max}, min pixel: {self.min}"

    @cached_property
    def max(self) -> np.ndarray:
        """The max pixel value of the ROI."""
        masked_img = self.circle_mask()
        return np.nanmax(masked_img)

    @cached_property
    def min(self) -> np.ndarray:
        """The min pixel value of the ROI."""
        masked_img = self.circle_mask()
        return np.nanmin(masked_img)


class RectangleROI(Rectangle):
    """Class that represents a rectangular ROI."""

    def __init__(self, array, width, height, angle, dist_from_center, phantom_center):
        y_shift = np.sin(np.deg2rad(angle)) * dist_from_center
        x_shift = np.cos(np.deg2rad(angle)) * dist_from_center
        center = Point(phantom_center.x + x_shift, phantom_center.y + y_shift)
        super().__init__(width, height, center, as_int=True)
        self._array = array

    def __repr__(self):
        return f"Rectangle ROI @ {self.center}; mean pixel: {self.pixel_value}"

    # TODO: See if I could use this somewhere
    # @classmethod
    # def from_regionprop(cls, regionprop: _RegionProperties, phan_center: Point):
    #     width = regionprop.bbox[3] - regionprop.bbox[1]
    #     height = regionprop.bbox[2] - regionprop.bbox[0]
    #     angle = np.rad2deg(np.arctan2((regionprop.centroid[0] - phan_center.y), (regionprop.centroid[1] - phan_center.x)))
    #     distance = phan_center.distance_to(Point(regionprop.centroid[1], regionprop.centroid[0]))
    #     return cls(regionprop.intensity_image, width=width, height=height,
    #                angle=angle, dist_from_center=distance, phantom_center=phan_center)

    @cached_property
    def pixel_array(self) -> np.ndarray:
        """The pixel array within the ROI."""
        return self._array[
            self.bl_corner.y : self.tr_corner.y, self.bl_corner.x : self.tr_corner.x
        ]

    @cached_property
    def pixel_value(self) -> float:
        """The pixel array within the ROI."""
        return float(np.mean(self.pixel_array))



"""I/O helper functions for pylinac."""
import os
import os.path as osp
import struct
import zipfile
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable, List, Tuple, Union, BinaryIO, Optional
from urllib.error import HTTPError, URLError
from urllib.request import urlretrieve, urlopen

import numpy as np
import pydicom
from tqdm import tqdm

from .profile import SingleProfile


def is_dicom(file: Union[str, Path]) -> bool:
    """Boolean specifying if file is a proper DICOM file.

    This function is a pared down version of read_preamble meant for a fast return.
    The file is read for a proper preamble ('DICM'), returning True if so,
    and False otherwise. This is a conservative approach.

    Parameters
    ----------
    file : str
        The path to the file.

    See Also
    --------
    pydicom.filereader.read_preamble
    pydicom.filereader.read_partial
    """
    with open(file, "rb") as fp:
        fp.read(0x80)
        prefix = fp.read(4)
        return prefix == b"DICM"


def is_dicom_image(file: Union[str, Path, BinaryIO]) -> bool:
    """Boolean specifying if file is a proper DICOM file with a image

    Parameters
    ----------
    file : str
        The path to the file.

    See Also
    --------
    pydicom.filereader.read_preamble
    pydicom.filereader.read_partial
    """
    result = False
    try:
        img = pydicom.dcmread(file, force=True)
        if "TransferSyntaxUID" not in img.file_meta:
            img.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
        img.pixel_array
        result = True
    except (AttributeError, TypeError, KeyError, struct.error):
        pass
    return result


def retrieve_dicom_file(file: Union[str, Path, BinaryIO]) -> pydicom.FileDataset:
    """Read and return the DICOM dataset.

    Parameters
    ----------
    file : str
        The path to the file.
    """
    ds = pydicom.dcmread(file, force=True)
    if "TransferSyntaxUID" not in ds.file_meta:
        ds.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
    return ds


class TemporaryZipDirectory(TemporaryDirectory):
    """Creates a temporary directory that unpacks a ZIP archive. Shockingly useful"""

    def __init__(self, zfile: Union[str, Path, BinaryIO]):
        """
        Parameters
        ----------
        zfile : str
            String that points to a ZIP archive.
        """
        super().__init__()
        zfiles = zipfile.ZipFile(zfile)
        zfiles.extractall(path=self.name)


def retrieve_filenames(
    directory: Union[str, Path], func: Optional[Callable] = None, recursive: bool = True, **kwargs
) -> List[str]:
    """Retrieve file names in a directory.

    Parameters
    ----------
    directory : str
        The directory to walk over recursively.
    func : function, None
        The function that validates if the file name should be kept.
        If None, no validation will be performed and all file names will be returned.
    recursive : bool
        Whether to search only the root directory.
    kwargs
        Additional arguments passed to the func parameter.
    """
    filenames = []
    if func is None:
        func = lambda x: True
    for pdir, _, files in os.walk(directory):
        for file in files:
            filename = osp.join(pdir, file)
            if func(filename, **kwargs):
                filenames.append(filename)
        if not recursive:
            break
    return filenames


def retrieve_demo_file(name: str, force: bool = False) -> Path:
    """Retrieve the demo file either by getting it from file or from a URL.

    If the file is already on disk it returns the file name. If the file isn't
    on disk, get the file from the URL and put it at the expected demo file location
    on disk for lazy loading next time.

    Parameters
    ----------
    name : str
        The suffix to the url (location within the S3 bucket) pointing to the demo file.
    """
    true_url = r"https://storage.googleapis.com/pylinac_demo_files/" + name
    demo_path = Path(__file__).parent.parent / "demo_files" / name
    # demo_file = osp.join(osp.dirname(osp.dirname(__file__)), "demo_files", name)
    demo_dir = demo_path.parent
    if not demo_dir.exists():
        os.makedirs(demo_dir)
    if force or not demo_path.exists():
        get_url(true_url, destination=demo_path)
    return demo_path


def is_url(url: str) -> bool:
    """Determine whether a given string is a valid URL.

    Parameters
    ----------
    url : str

    Returns
    -------
    bool
    """
    try:
        with urlopen(url) as r:
            return r.status == 200
    except:
        return False


def get_url(url: str, destination: Union[str, Path, None] = None, progress_bar: bool = True) -> str:
    """Download a URL to a local file.

    Parameters
    ----------
    url : str
        The URL to download.
    destination : str, None
        The destination of the file. If None is given the file is saved to a temporary directory.
    progress_bar : bool
        Whether to show a command-line progress bar while downloading.

    Returns
    -------
    filename : str
        The location of the downloaded file.

    Notes
    -----
    Progress bar use/example adapted from tqdm documentation: https://github.com/tqdm/tqdm
    """

    def my_hook(t):
        last_b = [0]

        def inner(b=1, bsize=1, tsize=None):
            if tsize is not None:
                t.total = tsize
            if b > 0:
                t.update((b - last_b[0]) * bsize)
            last_b[0] = b

        return inner

    try:
        if progress_bar:
            with tqdm(
                unit="B", unit_scale=True, miniters=1, desc=url.split("/")[-1]
            ) as t:
                filename, _ = urlretrieve(
                    url, filename=destination, reporthook=my_hook(t)
                )
        else:
            filename, _ = urlretrieve(url, filename=destination)
    except (HTTPError, URLError, ValueError) as e:
        raise e
    return filename


# this is easier with pandas, but I don't want that as a dependency at this point
class SNCProfiler:
    """Load a file from a Sun Nuclear Profiler device. This accepts .prs files."""

    def __init__(
        self,
        path: str,
        detector_row: int = 106,
        bias_row: int = 107,
        calibration_row: int = 108,
        data_row: int = -1,
        data_columns: slice = slice(5, 259),
    ):
        """
        Parameters
        ----------
        path : str
            Path to the .prs file.
        detector_row
        bias_row
        calibration_row
        data_row
        data_columns
            The range of columns that the data is in. Usually, there are some columns before and after the real data.
        """
        with open(path, encoding="cp437") as f:
            raw_data = f.read().splitlines()
            self.detectors = raw_data[detector_row].split("\t")[data_columns]
            self.bias = np.array(raw_data[bias_row].split("\t")[data_columns]).astype(
                float
            )
            self.calibration = np.array(
                raw_data[calibration_row].split("\t")[data_columns]
            ).astype(float)
            self.data = np.array(raw_data[data_row].split("\t")[data_columns]).astype(
                float
            )
            self.timetic = float(raw_data[bias_row].split("\t")[2])
            self.integrated_dose = self.calibration * (
                self.data - self.bias * self.timetic
            )

    def to_profiles(
        self, n_detectors_row: int = 63, **kwargs
    ) -> Tuple[SingleProfile, SingleProfile, SingleProfile, SingleProfile]:
        """Convert the SNC data to SingleProfiles. These can be analyzed directly or passed to other modules like flat/sym.

        Parameters
        ----------
        n_detectors_row : int
            The number of detectors in a given row. Note that they Y profile includes 2 extra detectors from the other 3.
        """
        def copy_cax_dose(array: np.ndarray, center_detector_idx: int = 31) -> np.ndarray:
            array = np.insert(array, center_detector_idx+1, array[center_detector_idx])
            array = np.insert(array, center_detector_idx-1, array[center_detector_idx])
            return array

        y_prof = SingleProfile(
            self.integrated_dose[n_detectors_row : 2 * n_detectors_row + 2], **kwargs
        )
        # for all but the y profile, we are missing detectors to the left and right of center because the center y-detector is too wide
        # for physical spacing purposes we have to fill those values in. we use the central value.
        x_prof = SingleProfile(copy_cax_dose(self.integrated_dose[:n_detectors_row]), **kwargs)
        pos_prof = SingleProfile(copy_cax_dose(
            self.integrated_dose[2 * n_detectors_row + 2 : 3 * n_detectors_row + 2]),
            **kwargs
        )
        neg_prof = SingleProfile(copy_cax_dose(
            self.integrated_dose[3 * n_detectors_row + 2 : 4 * n_detectors_row + 2]),
            **kwargs
        )
        return x_prof, y_prof, pos_prof, neg_prof



# -*- coding: utf-8 -*-
"""The VMAT module consists of the class VMAT, which is capable of loading an EPID DICOM Open field image and MLC field image and analyzing the
images according to the Varian RapidArc QA tests and procedures, specifically the Dose-Rate & Gantry-Speed (DRGS)
and Dose-Rate & MLC speed (DRMLC) tests.

Features:

* **Do both tests** - Pylinac can handle either DRGS or DRMLC tests.
* **Automatic offset correction** - Older VMAT tests had the ROIs offset, newer ones are centered. No worries, pylinac finds the ROIs automatically.
* **Automatic open/DMLC identification** - Pass in both images--don't worry about naming. Pylinac will automatically identify the right images.
"""
import dataclasses
import enum
import typing
import webbrowser
from dataclasses import dataclass
from io import BytesIO
from typing import Union, List, Tuple, Sequence, Optional, BinaryIO

import argue
import matplotlib.pyplot as plt
import numpy as np

from .core import image
from .core.geometry import Point, Rectangle
from .core.image import ImageLike
from .core.io import get_url, TemporaryZipDirectory, retrieve_demo_file
from .core.pdf import PylinacCanvas
from .core.profile import SingleProfile, Interpolation, Edge
from .core.utilities import ResultBase
from .settings import get_dicom_cmap


class ImageType(enum.Enum):
    """The image type options"""

    DMLC = "dmlc"  #:
    OPEN = "open"  #:
    PROFILE = "profile"  #:


@dataclass
class SegmentResult:
    """An individual segment/ROI result"""

    passed: bool  #:
    x_position_mm: float  #:
    r_corr: float  #:
    r_dev: float  #:
    center_x_y: float  #:


@dataclass
class VMATResult(ResultBase):
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    test_type: str  #:
    tolerance_percent: float  #:
    max_deviation_percent: float  #:
    abs_mean_deviation: float  #:
    passed: bool  #:
    segment_data: typing.Iterable[SegmentResult]  #:


class VMATBase:
    _url_suffix: str
    _result_header: str
    _result_short_header: str
    SEGMENT_X_POSITIONS_MM: Tuple
    dmlc_image: image.DicomImage
    open_image: image.DicomImage
    segments: List
    _tolerance: float

    def __init__(self, image_paths: Sequence[Union[str, BinaryIO]]):
        """
        Parameters
        ----------
        image_paths : iterable (list, tuple, etc)
            A sequence of paths to the image files.
        """
        if len(image_paths) != 2:
            raise ValueError("Exactly 2 images (open, DMLC) must be passed")
        image1, image2 = self._load_images(image_paths)
        image1, image2 = self._check_img_inversion(image1, image2)
        self._identify_images(image1, image2)
        self.segments = []
        self._tolerance = 0

    @classmethod
    def from_url(cls, url: str):
        """Load a ZIP archive from a URL.  Must follow the naming convention.

        Parameters
        ----------
        url : str
            Must point to a valid URL that is a ZIP archive of two VMAT images.
        """
        zfile = get_url(url)
        return cls.from_zip(zfile)

    @classmethod
    def from_zip(cls, path: str):
        """Load VMAT images from a ZIP file that contains both images. Must follow the naming convention.

        Parameters
        ----------
        path : str
            Path to the ZIP archive which holds the VMAT image files.
        """
        with TemporaryZipDirectory(path) as tmpzip:
            image_files = image.retrieve_image_files(tmpzip)
            return cls(image_paths=image_files)

    @classmethod
    def from_demo_images(cls):
        """Construct a VMAT instance using the demo images."""
        demo_file = retrieve_demo_file(name=cls._url_suffix)
        return cls.from_zip(demo_file)

    @argue.bounds(tolerance=(0, 8))
    def analyze(
        self, tolerance: Union[float, int] = 1.5, segment_size_mm: Tuple = (5, 100)
    ):
        """Analyze the open and DMLC field VMAT images, according to 1 of 2 possible tests.

        Parameters
        ----------
        tolerance : float, int, optional
            The tolerance of the sample deviations in percent. Default is 1.5.
            Must be between 0 and 8.
        segment_size_mm : tuple(int, int)
            The (width, height) of the ROI segments in mm.
        """
        self._tolerance = tolerance / 100

        """Analysis"""
        points = self._calculate_segment_centers()
        Segment._nominal_width_mm = segment_size_mm[0]
        Segment._nominal_height_mm = segment_size_mm[1]
        self._construct_segments(points)

    @staticmethod
    def _load_images(
        image_paths: Sequence[Union[str, BytesIO]]
    ) -> Tuple[ImageLike, ImageLike]:
        image1 = image.load(image_paths[0])
        image2 = image.load(image_paths[1])
        image1.ground()
        image2.ground()
        return image1, image2

    @staticmethod
    def _check_img_inversion(
        image1: ImageLike, image2: ImageLike
    ) -> Tuple[ImageLike, ImageLike]:
        """Check that the images are correctly inverted."""
        for image in [image1, image2]:
            image.check_inversion()
        return image1, image2

    def _identify_images(self, image1: ImageLike, image2: ImageLike):
        """Identify which image is the DMLC and which is the open field."""
        profile1, profile2 = self._median_profiles((image1, image2))
        field_profile1 = profile1.field_data()["field values"]
        field_profile2 = profile2.field_data()["field values"]
        if np.std(field_profile1) > np.std(field_profile2):
            self.dmlc_image = image1
            self.open_image = image2
        else:
            self.dmlc_image = image2
            self.open_image = image1

    def results(self) -> str:
        """A string of the summary of the analysis results.

        Returns
        -------
        str
            The results string showing the overall result and deviation statistics by segment.
        """
        if self.passed:
            passfail_str = "PASS"
        else:
            passfail_str = "FAIL"

        string = f"{self._result_header}\nTest Results (Tol. +/-{self._tolerance*100:2.2}%): {passfail_str}\n"

        string += f"Max Deviation: {self.max_r_deviation:2.3}%\nAbsolute Mean Deviation: {self.avg_abs_r_deviation:2.3}%"
        return string

    def results_data(self, as_dict=False) -> Union[VMATResult, dict]:
        """Present the results data and metadata as a dataclass or dict.
        The default return type is a dataclass."""
        segment_data = []
        for idx, segment in enumerate(self.segments):
            segment_data.append(
                SegmentResult(
                    passed=segment.passed,
                    r_corr=segment.r_corr,
                    r_dev=segment.r_dev,
                    center_x_y=segment.center.as_array(),
                    x_position_mm=self.SEGMENT_X_POSITIONS_MM[idx],
                )
            )
        data = VMATResult(
            test_type=self._result_header,
            tolerance_percent=self._tolerance * 100,
            max_deviation_percent=self.max_r_deviation,
            abs_mean_deviation=self.avg_abs_r_deviation,
            passed=self.passed,
            segment_data=segment_data,
        )

        if as_dict:
            return dataclasses.asdict(data)
        return data

    def _calculate_segment_centers(self) -> List[Point]:
        """Construct the center points of the segments based on the field center and known x-offsets."""
        points = []
        dmlc_prof, _ = self._median_profiles((self.dmlc_image, self.open_image))
        x_field_center = dmlc_prof.beam_center()["index (rounded)"]
        for x_offset_mm in self.SEGMENT_X_POSITIONS_MM:
            y = self.open_image.center.y
            x_offset_pixels = x_offset_mm * self.open_image.dpmm
            x = x_field_center + x_offset_pixels
            points.append(Point(x, y))
        return points

    def _construct_segments(self, points: List[Point]):
        for point in points:
            segment = Segment(point, self.open_image, self.dmlc_image, self._tolerance)
            self.segments.append(segment)
        # post-analysis to update R_corr values
        self._update_r_corrs()

    def _update_r_corrs(self):
        """After the Segment constructions, the R_corr must be set for each segment."""
        avg_r_corr = np.array([segment.r_corr for segment in self.segments]).mean()
        for segment in self.segments:
            segment.r_dev = ((segment.r_corr / avg_r_corr) * 100) - 100

    @property
    def passed(self) -> bool:
        return all(segment.passed for segment in self.segments)

    @property
    def r_devs(self) -> np.ndarray:
        """Return the deviations of all segments as an array."""
        return np.array([segment.r_dev for segment in self.segments])

    @property
    def avg_abs_r_deviation(self) -> float:
        """Return the average of the absolute R_deviation values."""
        return np.abs(self.r_devs).mean()

    @property
    def avg_r_deviation(self) -> float:
        """Return the average of the R_deviation values, including the sign."""
        return self.r_devs.mean()

    @property
    def max_r_deviation(self) -> float:
        """Return the value of the maximum R_deviation segment."""
        return np.max(np.abs(self.r_devs))

    def plot_analyzed_image(self, show: bool = True, **plt_kwargs: dict):
        """Plot the analyzed images. Shows the open and dmlc images with the segments drawn; also plots the median
        profiles of the two images for visual comparison.

        Parameters
        ----------
        show : bool
            Whether to actually show the image.
        plt_kwargs : dict
            Keyword args passed to the plt.subplots() method. Allows one to set things like figure size.
        """
        fig, axes = plt.subplots(ncols=3, sharex=True, **plt_kwargs)
        subimages = (ImageType.OPEN, ImageType.DMLC, ImageType.PROFILE)
        titles = ("Open", "DMLC", "Median Profiles")
        for subimage, axis, title in zip(subimages, axes, titles):
            self._plot_analyzed_subimage(subimage=subimage, ax=axis, show=False)
            axis.set_title(title)
        axis.set_ylabel("Normalized Response")
        axis.legend(loc="lower center")

        if show:
            plt.tight_layout(h_pad=1.5)
            plt.show()

    def _save_analyzed_subimage(
        self, filename: Union[str, BytesIO], subimage: ImageType, **kwargs
    ):
        """Save the analyzed images as a png file.

        Parameters
        ----------
        filename : str, file-object
            Where to save the file to.
        kwargs
            Passed to matplotlib.
        """
        self._plot_analyzed_subimage(subimage=subimage, show=False)
        plt.savefig(filename, **kwargs)

    def _plot_analyzed_subimage(
        self, subimage: ImageType, show: bool = True, ax: Optional[plt.Axes] = None
    ):
        """Plot an individual piece of the VMAT analysis.

        Parameters
        ----------
        subimage : str
            Specifies which image to plot.
        show : bool
            Whether to actually plot the image.
        ax : matplotlib Axes, None
            If None (default), creates a new figure to plot to, otherwise plots to the given axes.
        """
        plt.ioff()
        if ax is None:
            fig, ax = plt.subplots()

        # plot DMLC or OPEN image
        if subimage in (ImageType.DMLC, ImageType.OPEN):
            if subimage == ImageType.DMLC:
                img = self.dmlc_image
            elif subimage == ImageType.OPEN:
                img = self.open_image
            ax.imshow(img, cmap=get_dicom_cmap())
            self._draw_segments(ax)
            plt.sca(ax)
            plt.axis("off")
            plt.tight_layout()

        # plot profile
        elif subimage == ImageType.PROFILE:
            dmlc_prof, open_prof = self._median_profiles(
                (self.dmlc_image, self.open_image)
            )
            ax.plot(dmlc_prof.values, label="DMLC")
            ax.plot(open_prof.values, label="Open")
            ax.autoscale(axis="x", tight=True)
            ax.legend(loc=8, fontsize="large")
            ax.grid()

        if show:
            plt.show()

    def _draw_segments(self, axis: plt.Axes):
        """Draw the segments onto a plot.

        Parameters
        ----------
        axis : matplotlib.axes.Axes
            The plot to draw the objects on.
        """
        for segment in self.segments:
            color = segment.get_bg_color()
            segment.plot2axes(axis, edgecolor=color)

    @staticmethod
    def _median_profiles(images) -> Tuple[SingleProfile, SingleProfile]:
        """Return two median profiles from the open and dmlc image. For visual comparison."""
        profile1 = SingleProfile(
            np.mean(images[0], axis=0),
            interpolation=Interpolation.NONE,
            edge_detection_method=Edge.INFLECTION_DERIVATIVE,
        )
        profile1.stretch()
        profile2 = SingleProfile(
            np.mean(images[1], axis=0),
            interpolation=Interpolation.NONE,
            edge_detection_method=Edge.INFLECTION_DERIVATIVE,
        )
        profile2.stretch()

        # normalize the profiles to approximately the same value
        norm_val = np.percentile(profile1.values, 90)
        profile1.normalize(norm_val)
        norm_val = np.percentile(profile2.values, 90)
        profile2.normalize(norm_val)

        return profile1, profile2

    def publish_pdf(
        self,
        filename: str,
        notes: str = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ):
        """Publish (print) a PDF containing the analysis, images, and quantitative results.

        Parameters
        ----------
        filename : (str, file-like object}
            The file to write the results to.
        notes : str, list of strings
            Text; if str, prints single line.
            If list of strings, each list item is printed on its own line.
        open_file : bool
            Whether to open the file using the default program after creation.
        metadata : dict
            Extra data to be passed and shown in the PDF. The key and value will be shown with a colon.
            E.g. passing {'Author': 'James', 'Unit': 'TrueBeam'} would result in text in the PDF like:
            --------------
            Author: James
            Unit: TrueBeam
            --------------
        """
        canvas = PylinacCanvas(
            filename=filename,
            page_title=f"{self._result_short_header} VMAT Analysis",
            metadata=metadata,
        )
        for y, x, width, img in zip(
            (9, 9, -2),
            (1, 11, 3),
            (9, 9, 14),
            (ImageType.OPEN, ImageType.DMLC, ImageType.PROFILE),
        ):
            data = BytesIO()
            self._save_analyzed_subimage(data, subimage=img)
            canvas.add_image(data, location=(x, y), dimensions=(width, 18))
            # canvas.add_text(text=f"{img} Image", location=(x + 2, y + 10), font_size=18)
        canvas.add_text(text="Open Image", location=(4, 22), font_size=18)
        canvas.add_text(text=f"{self.open_image.base_path}", location=(4, 21.5))
        canvas.add_text(text="DMLC Image", location=(14, 22), font_size=18)
        canvas.add_text(text=f"{self.dmlc_image.base_path}", location=(14, 21.5))
        canvas.add_text(text="Median profiles", location=(8, 12), font_size=18)
        text = [
            f"{self._result_header} VMAT results:",
            f"Source-to-Image Distance (mm): {self.open_image.sid:2.0f}",
            f"Tolerance (%): {self._tolerance*100:2.1f}",
            f"Absolute mean deviation (%): {self.avg_abs_r_deviation:2.2f}",
            f"Maximum deviation (%): {self.max_r_deviation:2.2f}",
        ]
        canvas.add_text(text=text, location=(10, 25.5))
        if notes is not None:
            canvas.add_text(text="Notes:", location=(1, 5.5), font_size=14)
            canvas.add_text(text=notes, location=(1, 5))

        canvas.finish()

        if open_file:
            webbrowser.open(filename)


class DRGS(VMATBase):
    """Class representing a Dose-Rate, Gantry-speed VMAT test. Will accept, analyze, and return the results."""

    _url_suffix = "drgs.zip"
    _result_header = "Dose Rate & Gantry Speed"
    _result_short_header = "DR/GS"
    SEGMENT_X_POSITIONS_MM = (-60, -40, -20, 0, 20, 40, 60)

    @staticmethod
    def run_demo():
        """Run the demo for the Dose Rate & Gantry Speed test."""
        vmat = DRGS.from_demo_images()
        vmat.analyze()  # old images (rev1, not new rev2's), which are offset
        print(vmat.results())
        vmat.plot_analyzed_image()


class DRMLC(VMATBase):
    """Class representing a Dose-Rate, MLC speed VMAT test. Will accept, analyze, and return the results."""

    _url_suffix = "drmlc.zip"
    _result_header = "Dose Rate & MLC Speed"
    _result_short_header = "DR/MLCS"
    SEGMENT_X_POSITIONS_MM = (-45, -15, 15, 45)

    @staticmethod
    def run_demo():
        """Run the demo for the MLC leaf speed test."""
        vmat = DRMLC.from_demo_images()
        vmat.analyze()
        print(vmat.results())
        vmat.plot_analyzed_image()


class Segment(Rectangle):
    """A class for holding and analyzing segment data of VMAT tests.

    For VMAT tests, there are either 4 or 7 'segments', which represents a section of the image that received
    radiation under the same conditions.

    Attributes
    ----------
    r_dev : float
            The reading deviation (R_dev) from the average readings of all the segments. See RTD for equation info.
    r_corr : float
        The corrected reading (R_corr) of the pixel values. See RTD for explanation and equation info.
    passed : boolean
        Specifies where the segment reading deviation was under tolerance.
    """

    # width of the segment (i.e. parallel to MLC motion) in pixels under reference conditions
    _nominal_width_mm: int
    _nominal_height_mm: int

    def __init__(
        self,
        center_point: Point,
        open_image: image.DicomImage,
        dmlc_image: image.DicomImage,
        tolerance: Union[float, int],
    ):
        self.r_dev: float = 0.0  # is assigned after all segments constructed
        self._tolerance = tolerance
        self._open_image = open_image
        self._dmlc_image = dmlc_image
        width = self._nominal_width_mm * dmlc_image.dpmm
        height = self._nominal_height_mm * dmlc_image.dpmm
        super().__init__(width, height, center=center_point, as_int=True)

    @property
    def r_corr(self) -> float:
        """Return the ratio of the mean pixel values of DMLC/OPEN images."""
        dmlc_value = self._dmlc_image.array[
            self.bl_corner.y : self.bl_corner.y + self.height,
            self.bl_corner.x : self.bl_corner.x + self.width,
        ].mean()
        open_value = self._open_image.array[
            self.bl_corner.y : self.bl_corner.y + self.height,
            self.bl_corner.x : self.bl_corner.x + self.width,
        ].mean()
        ratio = (dmlc_value / open_value) * 100
        return ratio

    @property
    def passed(self) -> bool:
        """Return whether the segment passed or failed."""
        return abs(self.r_dev) < self._tolerance * 100

    def get_bg_color(self) -> str:
        """Get the background color of the segment when plotted, based on the pass/fail status."""
        return "blue" if self.passed else "red"



"""
The TG-51 module contains a number of helper functions and classes that can calculate parameters for performing the
TG-51 absolute linac dose calibration although there are some modifications from the original TG-51. The modifications 
include updated kQ and kecal values from Muir and Rogers' set of papers.
Functions include all relevant calculations for TG-51 including PDDx, kQ,
Dref, and chamber reading corrections. Where Muir & Rogers' values/equations are used they are specified in the documentation.

Classes include photon and electron calibrations using cylindrical chambers. Pass all the relevant raw measurements
and the class will compute all corrections and corrected readings and dose at 10cm and dmax/dref.
"""
import webbrowser
from datetime import datetime
from typing import Optional

import argue
import numpy as np

from ..core.pdf import PylinacCanvas
from ..core.typing import NumberOrArray
from ..core.utilities import Structure

MIN_TEMP = 15
MAX_TEMP = 35
MIN_PRESSURE = 90
MAX_PRESSURE = 115
MIN_PION = 1
MAX_PION = 1.05
MIN_PTP = 0.9
MAX_PTP = 1.1
MIN_PELEC = 0.98
MAX_PELEC = 1.02
MIN_PPOL = 0.98
MAX_PPOL = 1.02


KQ_PHOTONS = {
    # Exradin
    "A12": {
        "a": 1.0146,
        "b": 0.777e-3,
        "c": -1.666e-5,
        "a'": 2.6402,
        "b'": -7.2304,
        "c'": 10.7573,
        "d'": -5.4294,
    },
    "A19": {
        "a": 0.9934,
        "b": 1.384e-3,
        "c": -2.125e-5,
        "a'": 3.0907,
        "b'": -9.1930,
        "c'": 13.5957,
        "d'": -6.7969,
    },
    "A2": {
        "a": 0.9819,
        "b": 1.609e-3,
        "c": -2.184e-5,
        "a'": 2.8458,
        "b'": -8.1619,
        "c'": 12.1411,
        "d'": -6.1041,
    },
    "T2": {
        "a": 1.0173,
        "b": 0.854e-3,
        "c": -1.941e-5,
        "a'": 3.3433,
        "b'": -10.2649,
        "c'": 15.1247,
        "d'": -7.5415,
    },
    "A12S": {
        "a": 0.9692,
        "b": 1.974e-3,
        "c": -2.448e-5,
        "a'": 2.9597,
        "b'": -8.6777,
        "c'": 12.9155,
        "d'": -6.4903,
    },
    "A18": {
        "a": 0.9944,
        "b": 1.286e-3,
        "c": -1.980e-5,
        "a'": 2.5167,
        "b'": -6.7567,
        "c'": 10.1519,
        "d'": -5.1709,
    },
    "A1": {
        "a": 1.0029,
        "b": 1.023e-3,
        "c": -1.803e-5,
        "a'": 2.0848,
        "b'": -4.9174,
        "c'": 7.5446,
        "d'": -3.9441,
    },
    "T1": {
        "a": 1.0552,
        "b": -0.196e-3,
        "c": -1.275e-5,
        "a'": 2.8060,
        "b'": -7.9273,
        "c'": 11.7541,
        "d'": -5.9263,
    },
    "A1SL": {
        "a": 0.9896,
        "b": 1.410e-3,
        "c": -2.049e-5,
        "a'": 2.8029,
        "b'": -7.9648,
        "c'": 11.8445,
        "d'": -5.9568,
    },
    "A14": {
        "a": 0.9285,
        "b": 2.706e-3,
        "c": -2.599e-5,
        "a'": 5.4677,
        "b'": -19.1795,
        "c'": 27.4542,
        "d'": -13.1336,
    },
    "T14": {
        "a": 0.9622,
        "b": 2.009e-3,
        "c": -2.401e-5,
        "a'": 4.9690,
        "b'": -17.1074,
        "c'": 24.6292,
        "d'": -11.8877,
    },
    "A14SL": {
        "a": 0.9017,
        "b": 3.454e-3,
        "c": -3.083e-5,
        "a'": 5.1205,
        "b'": -17.7884,
        "c'": 25.6123,
        "d'": -12.3232,
    },
    "A16": {
        "a": 0.8367,
        "b": 4.987e-3,
        "c": -3.877e-5,
        "a'": 6.0571,
        "b'": -21.7829,
        "c'": 31.2289,
        "d'": -14.9168,
    },
    # PTW
    "30010": {
        "a": 1.0093,
        "b": 0.926e-3,
        "c": -1.771e-5,
        "a'": 2.5318,
        "b'": -6.7948,
        "c'": 10.1779,
        "d'": -5.1746,
    },
    "30011": {
        "a": 0.9676,
        "b": 2.061e-3,
        "c": -2.528e-5,
        "a'": 2.9044,
        "b'": -8.4576,
        "c'": 12.6339,
        "d'": -6.3742,
    },
    "30012": {
        "a": 0.9537,
        "b": 2.440e-3,
        "c": -2.750e-5,
        "a'": 3.2836,
        "b'": -10.0610,
        "c'": 14.8867,
        "d'": -7.4212,
    },
    "30013": {
        "a": 0.9652,
        "b": 2.141e-3,
        "c": -2.623e-5,
        "a'": 3.2012,
        "b'": -9.7211,
        "c'": 14.4211,
        "d'": -7.2184,
    },
    "31010": {
        "a": 0.9590,
        "b": 2.265e-3,
        "c": -2.684e-5,
        "a'": 3.1578,
        "b'": -9.5422,
        "c'": 14.1676,
        "d'": -7.0964,
    },
    "31016": {
        "a": 1.0085,
        "b": 1.028e-3,
        "c": -1.968e-5,
        "a'": 2.9524,
        "b'": -8.6054,
        "c'": 12.7757,
        "d'": -6.4265,
    },
    "31014": {
        "a": 1.0071,
        "b": 1.048e-3,
        "c": -1.967e-5,
        "a'": 3.0178,
        "b'": -8.8735,
        "c'": 13.1372,
        "d'": -6.5867,
    },
    # IBA
    "CC25": {
        "a": 0.9551,
        "b": 2.353e-3,
        "c": -2.687e-5,
        "a'": 2.4567,
        "b'": -6.5932,
        "c'": 10.0471,
        "d'": -5.1775,
    },
    "CC13": {
        "a": 0.9515,
        "b": 2.455e-3,
        "c": -2.768e-5,
        "a'": 3.1982,
        "b'": -9.7182,
        "c'": 14.4210,
        "d'": -7.2121,
    },
    "CC08": {
        "a": 0.9430,
        "b": 2.637e-3,
        "c": -2.884e-5,
        "a'": 3.7328,
        "b'": -11.9800,
        "c'": 17.5884,
        "d'": -8.6843,
    },
    "CC04": {
        "a": 0.9714,
        "b": 1.938e-3,
        "c": -2.432e-5,
        "a'": 3.0054,
        "b'": -8.8633,
        "c'": 13.1704,
        "d'": -6.6075,
    },
    "CC01": {
        "a": 0.9116,
        "b": 3.358e-3,
        "c": -3.177e-5,
        "a'": 4.3376,
        "b'": -14.4935,
        "c'": 21.0293,
        "d'": -10.2208,
    },
    "FC65-G": {
        "a": 0.9708,
        "b": 1.972e-3,
        "c": -2.480e-5,
        "a'": 3.3221,
        "b'": -10.2012,
        "c'": 15.0497,
        "d'": -7.4872,
    },
    "FC65-P": {
        "a": 0.9828,
        "b": 1.664e-3,
        "c": -2.296e-5,
        "a'": 3.0872,
        "b'": -9.1919,
        "c'": 13.6137,
        "d'": -6.8118,
    },
    "FC23-C": {
        "a": 0.9820,
        "b": 1.579e-3,
        "c": -2.166e-5,
        "a'": 3.0511,
        "b'": -9.0243,
        "c'": 13.3378,
        "d'": -6.6559,
    },
    # Other
    "NE2581": {
        "a": 1.0318,
        "b": 0.488e-3,
        "c": -1.731e-5,
        "a'": 2.9190,
        "b'": -8.4561,
        "c'": 12.5690,
        "d'": -6.3468,
    },
    "NE2571": {
        "a": 0.9882,
        "b": 1.486e-3,
        "c": -2.140e-5,
        "a'": 2.2328,
        "b'": -5.5779,
        "c'": 8.5325,
        "d'": -4.4352,
    },
    "NE2561": {
        "a": 1.0200,
        "b": 0.596e-3,
        "c": -1.551e-5,
        "a'": 2.4235,
        "b'": -6.3179,
        "c'": 9.4737,
        "d'": -4.8307,
    },
    "PR06C/G": {
        "a": 0.9519,
        "b": 2.432e-3,
        "c": -2.704e-5,
        "a'": 2.9110,
        "b'": -8.4916,
        "c'": 12.6817,
        "d'": -6.3874,
    },
}

KQ_ELECTRONS = {
    # Exradin
    "A12": {"kQ,ecal": 0.907, "a": 0.965, "b": 0.119, "c": 0.607},
    "A19": {"kQ,ecal": 0.904, "a": 0.957, "b": 0.119, "c": 0.505},
    "A12S": {"kQ,ecal": 0.907, "a": 0.937, "b": 0.136, "c": 0.378},
    "A18": {"kQ,ecal": 0.914, "a": 0.352, "b": 0.711, "c": 0.046},
    "A1SL": {"kQ,ecal": 0.914, "a": 0.205, "b": 0.854, "c": 0.036},
    # PTW
    "30010": {"kQ,ecal": 0.904, "a": 0.980, "b": 0.119, "c": 0.891},
    "30011": {"kQ,ecal": 0.901, "a": 0.976, "b": 0.120, "c": 0.793},
    "30012": {"kQ,ecal": 0.908, "a": 0.972, "b": 0.121, "c": 0.728},
    "30013": {"kQ,ecal": 0.901, "a": 0.978, "b": 0.112, "c": 0.816},
    "31013": {"kQ,ecal": 0.902, "a": 0.945, "b": 0.133, "c": 0.441},
    # IBA
    "FC65-G": {"kQ,ecal": 0.904, "a": 0.971, "b": 0.113, "c": 0.680},
    "FC65-P": {"kQ,ecal": 0.902, "a": 0.973, "b": 0.110, "c": 0.692},
    "FC23-C": {"kQ,ecal": 0.904, "a": 0.971, "b": 0.097, "c": 0.591},
    "CC25": {"kQ,ecal": 0.904, "a": 0.964, "b": 0.105, "c": 0.539},
    "CC13": {"kQ,ecal": 0.904, "a": 0.926, "b": 0.129, "c": 0.279},
    # Other
    "PR06C/G": {"kQ,ecal": 0.906, "a": 0.972, "b": 0.122, "c": 0.729},
    "NE2571": {"kQ,ecal": 0.903, "a": 0.977, "b": 0.117, "c": 0.817},
    "NE2611": {"kQ,ecal": 0.896, "a": 0.979, "b": 0.120, "c": 0.875},
}

LEAD_OPTIONS = {"None": None, "30cm": "30cm", "50cm": "50cm"}


def mmHg2kPa(mmHg: float) -> float:
    """Utility function to convert from mmHg to kPa."""
    return mmHg * 101.33 / 760


def mbar2kPa(mbar: float) -> float:
    """Utility function to convert from millibars to kPa."""
    return mbar / 10


def fahrenheit2celsius(f: float) -> float:
    """Utility function to convert from Fahrenheit to Celsius."""
    return (f - 32) * 5 / 9


@argue.bounds(pdd2010=(0.5, 1))
def tpr2010_from_pdd2010(*, pdd2010: float) -> float:
    """Calculate TPR20,10 from PDD20,10. From TRS-398 pg 62 and Followill et al 1998 eqn 1."""
    return 1.2661 * pdd2010 - 0.0595


def p_tp(*, temp: float, press: float) -> float:
    """Calculate the temperature & pressure correction.

    Parameters
    ----------
    temp : float (17-27)
        The temperature in degrees Celsius.
    press : float (91-111)
        The value of pressure in kPa. Can be converted from mmHg and mbar;
        see :func:`~pylinac.calibration.tg51.mmHg2kPa` and :func:`~pylinac.calibration.tg51.mbar2kPa`.
    """
    argue.verify_bounds(
        temp,
        bounds=(MIN_TEMP, MAX_TEMP),
        message="Temperature {:2.2f} out of range. Did you use Fahrenheit? Consider using the utility function fahrenheit2celsius()",
    )
    argue.verify_bounds(
        press,
        bounds=(MIN_PRESSURE, MAX_PRESSURE),
        message="Pressure {:2.2f} out of range. Did you use kPa? Consider using the utility functions mmHg2kPa() or mbar2kPa()",
    )
    return ((273.2 + temp) / 295.2) * (101.33 / press)


def p_pol(*, m_reference: NumberOrArray, m_opposite: NumberOrArray) -> float:
    """Calculate the polarity correction.

    Parameters
    ----------
    m_reference : number, array
        The readings of the ion chamber at the reference polarity and voltage.
    m_opposite : number, array
        The readings of the ion chamber at the polarity opposite the reference. The sign does not make a difference.

    Raises
    ------
    BoundsError if calculated Ppol is >1% from 1.0.
    """
    mref_avg = np.mean(m_reference)
    mopp_avg = np.mean(m_opposite)
    polarity = (abs(mref_avg) + abs(mopp_avg)) / abs(2 * mref_avg)
    argue.verify_bounds(
        polarity,
        bounds=(MIN_PPOL, MAX_PPOL),
        message="Polarity correction {:2.2f} out of range (+/-2%). Verify inputs",
    )
    return float(polarity)


def p_ion(
    *,
    voltage_reference: int,
    voltage_reduced: int,
    m_reference: NumberOrArray,
    m_reduced: NumberOrArray,
) -> float:
    """Calculate the ion chamber collection correction.

    Parameters
    ----------
    voltage_reference : int
        The "high" voltage; same as the TG51 measurement voltage.
    voltage_reduced : int
        The "low" voltage; usually half of the high voltage.
    m_reference : float, iterable
        The readings of the ion chamber at the "high" voltage.
    m_reduced : float, iterable
        The readings of the ion chamber at the "low" voltage.

    Raises
    ------
    BoundsError if calculated Pion is outside the range 1.00-1.05.
    """
    ion = (1 - voltage_reference / voltage_reduced) / (
        np.mean(m_reference) / np.mean(m_reduced) - voltage_reference / voltage_reduced
    )
    argue.verify_bounds(
        ion,
        bounds=(MIN_PION, MAX_PION),
        message="Pion out of range (1.00-1.05). Check inputs or chamber",
    )
    return float(ion)


def d_ref(*, i_50: float) -> float:
    """Calculate the dref of an electron beam based on the I50 depth.

    Parameters
    ----------
    i_50 : float
        The value of I50 in cm.
    """
    argue.verify_bounds(i_50, bounds=argue.POSITIVE, message="i50 should be positive")
    r50 = r_50(i_50=i_50)
    return 0.6 * r50 - 0.1


def r_50(*, i_50: float) -> float:
    """Calculate the R50 depth of an electron beam based on the I50 depth.

    Parameters
    ----------
    i_50 : float
        The value of I50 in cm.
    """
    argue.verify_bounds(i_50, bounds=argue.POSITIVE, message="i50 should be positive")
    if i_50 < 10:
        r50 = 1.029 * i_50 - 0.06
    else:
        r50 = 1.59 * i_50 - 0.37
    return r50


def kp_r50(*, r_50: float) -> float:
    """Calculate k'R50 for Farmer-like chambers.

    Parameters
    ----------
    r_50 : float (2-9)
        The R50 value in cm.
    """
    argue.verify_bounds(r_50, bounds=(2, 9))
    return 0.9905 + 0.071 * np.exp(-r_50 / 3.67)


def pq_gr(*, m_dref_plus: NumberOrArray, m_dref: NumberOrArray) -> float:
    """Calculate PQ_gradient for a cylindrical chamber.

    Parameters
    ----------
    m_dref_plus : float, iterable
        The readings of the ion chamber at dref + 0.5rcav.
    m_dref : float, iterable
        The readings of the ion chamber at dref.
    """
    return float(np.mean(m_dref_plus) / np.mean(m_dref))


def m_corrected(
    *,
    p_ion: float,
    p_tp: float,
    p_elec: float,
    p_pol: float,
    m_reference: NumberOrArray,
) -> float:
    """Calculate M_corrected, the ion chamber reading with all corrections applied.

    Parameters
    ----------
    p_ion : float (1.00-1.05)
        The ion collection correction.
    p_tp : float (0.92-1.08)
        The temperature & pressure correction.
    p_elec : float (0.98-1.02)
        The electrometer correction.
    p_pol : float (0.98-1.02)
        The polarity correction.
    m_reference : float, iterable
        The raw ion chamber reading(s).

    Returns
    -------
    float
    """
    argue.verify_bounds(p_ion, bounds=(MIN_PION, MAX_PION))
    argue.verify_bounds(p_tp, bounds=(MIN_PTP, MAX_PTP))
    argue.verify_bounds(p_elec, bounds=(MIN_PELEC, MAX_PELEC))
    argue.verify_bounds(p_pol, bounds=(MIN_PPOL, MAX_PPOL))
    return float(p_ion * p_tp * p_elec * p_pol * np.mean(m_reference))


@argue.bounds(pdd=(62.7, 89.0))
@argue.options(lead_foil=LEAD_OPTIONS.values())
def pddx(*, pdd: float, energy: int, lead_foil: Optional[str] = None) -> float:
    """Calculate PDDx based on the PDD.

    Parameters
    ----------
    pdd : {>62.7, <89.0}
        The measured PDD. If lead foil was used, this assumes the pdd as measured with the lead in place.
    energy : int
        The nominal energy in MV.
    lead_foil : {None, '30cm', '50cm'}
        Applicable only for energies >10MV.
        Whether a lead foil was used to acquire the pdd.
        Use ``None`` if no lead foil was used and the interim equation should be used. This is the default
        Use ``50cm`` if the lead foil was set to 50cm from the phantom surface.
        Use ``30cm`` if the lead foil was set to 30cm from the phantom surface.
    """
    if energy < 10:
        return pdd
    elif energy >= 10:
        if lead_foil is None:
            if pdd <= 75:
                return pdd
            elif 75 < pdd <= 89:
                return 1.267 * pdd - 20
            else:
                raise ValueError(f"PDD value of {pdd} was outside the bound of 89%")
        elif lead_foil == LEAD_OPTIONS["50cm"]:
            if pdd < 73:
                return pdd
            else:
                return (0.8905 + 0.0015 * pdd) * pdd
        elif lead_foil == LEAD_OPTIONS["30cm"]:
            if pdd < 71:
                return pdd
            else:
                return (0.8116 + 0.00264 * pdd) * pdd


@argue.bounds(pddx=(63.0, 86.0))
@argue.options(chamber=KQ_PHOTONS.keys())
def kq_photon_pddx(*, chamber: str, pddx: float) -> float:
    """Calculate kQ based on the chamber and clinical measurements of PDD(10)x. This will calculate kQ for photons
    for *CYLINDRICAL* chambers only.

    Parameters
    ----------
    chamber : str
        The chamber of the chamber. Valid values are those listed in
        Table III of Muir and Rogers and Table I of the TG-51 Addendum.
    pddx : {>63.0, <86.0}
        The **PHOTON-ONLY** PDD measurement at 10cm depth for a 10x10cm2 field.

        .. note:: Use the :func:`~pylinac.calibration.tg51.pddx` function to convert PDD to PDDx as needed.

        .. note:: Muir and Rogers state limits of 0.627 - 0.861. The TG-51 addendum states them as 0.63 and 0.86.
                  The TG-51 addendum limits are used here.
    """
    ch = KQ_PHOTONS[chamber]
    return ch["a"] + ch["b"] * pddx + ch["c"] * (pddx**2)


@argue.bounds(tpr=(0.623, 0.805))
@argue.options(chamber=KQ_PHOTONS.keys())
def kq_photon_tpr(*, chamber: str, tpr: float) -> float:
    """Calculate kQ based on the chamber and clinical measurements of TPR20,10. This will calculate kQ for photons
    for *CYLINDRICAL* chambers only.

    Parameters
    ----------
    chamber : str
        The chamber of the chamber. Valid values are those listed in
        Table III of Muir and Rogers and Table I of the TG-51 Addendum.
    tpr : {>0.630, <0.860}
        The TPR(20,10) value.

        .. note::
         Use the :func:`~pylinac.calibration.tg51.tpr2010_from_pdd2010` function to convert from PDD without needing to take TPR measurements.
    """
    ch = KQ_PHOTONS[chamber]
    return ch["a'"] + ch["b'"] * tpr + ch["c'"] * (tpr**2) + ch["d'"] * (tpr**3)


@argue.options(chamber=KQ_ELECTRONS.keys())
def kq_electron(*, chamber: str, r_50: float) -> float:
    """Calculate kQ based on the chamber and clinical measurements. This will calculate kQ for electrons
    for *CYLINDRICAL* chambers only according to Muir & Rogers.

    Parameters
    ----------
    chamber : str
        The chamber of the chamber. Valid values are those listed in
        Tables VI and VII of Muir and Rogers 2014.
    r_50 : float
        The R50 value in cm of an electron beam.
    """
    ch = KQ_ELECTRONS[chamber]
    return (ch["a"] + ch["b"] * r_50 ** -ch["c"]) * ch["kQ,ecal"]


class TG51Base(Structure):
    @property
    def p_tp(self) -> float:
        """Temperature/Pressure correction."""
        return p_tp(temp=self.temp, press=self.press)

    @property
    def p_ion(self) -> float:
        """Ionization collection correction."""
        return p_ion(
            voltage_reference=self.voltage_reference,
            voltage_reduced=self.voltage_reduced,
            m_reference=self.m_reference,
            m_reduced=self.m_reduced,
        )

    @property
    def p_pol(self) -> float:
        """Polarity correction."""
        return p_pol(m_reference=self.m_reference, m_opposite=self.m_opposite)

    @property
    def m_corrected(self) -> float:
        """Corrected chamber reading."""
        return m_corrected(
            p_ion=self.p_ion,
            p_tp=self.p_tp,
            p_elec=self.p_elec,
            p_pol=self.p_pol,
            m_reference=self.m_reference,
        )

    @property
    def m_corrected_adjustment(self) -> float:
        """Corrected chamber reading after adjusting the output."""
        if self.m_reference_adjusted is not None:
            return m_corrected(
                p_ion=self.p_ion,
                p_tp=self.p_tp,
                p_elec=self.p_elec,
                p_pol=self.p_pol,
                m_reference=self.m_reference_adjusted,
            )

    @property
    def output_was_adjusted(self) -> float:
        """Boolean specifiying if output was adjusted."""
        return self.m_reference_adjusted is not None


class TG51Photon(TG51Base):
    """Class for calculating absolute dose to water using a cylindrical chamber in a photon beam.

    Parameters
    ----------
    institution : str
        Institution name.
    physicist : str
        Physicist performing calibration.
    unit : str
        Unit name; e.g. TrueBeam1.
    measurement_date : str
        Date of measurement. E.g. 10/22/2018.
    temp : float
        The temperature in Celsius. Use :func:`~pylinac.calibration.tg51.fahrenheit2celsius` to convert if necessary.
    press : float
        The value of pressure in kPa. Can be converted from mmHg and mbar; see :func:`~pylinac.calibration.tg51.mmHg2kPa` and :func:`~pylinac.calibration.tg51.mbar2kPa`.
    energy : float
        Nominal energy of the beam in MV.
    chamber : str
        Chamber model. Must be one of the listed chambers in TG-51 Addendum.
    n_dw : float
        NDW value in Gy/nC.
    p_elec : float
        Electrometer correction factor; given by the calibration laboratory.
    measured_pdd10 : float
        The measured value of PDD(10); will be converted to PDDx(10) and used for calculating kq.
    lead_foil : {None, '50cm', '30cm'}
        Whether a lead foil was used to acquire PDD(10)x and where its position was. Used to calculate kq.
    clinical_pdd10 : float
        The PDD used to correct the dose at 10cm back to dmax. Usually the TPS PDD(10) value.
    voltage_reference : int
        Reference voltage; i.e. voltage when taking the calibration measurement.
    voltage_reduced : int
        Reduced voltage; usually half of the reference voltage.
    m_reference : float, tuple
        Ion chamber reading(s) at the reference voltage.
    m_opposite : float, tuple
        Ion chamber reading(s) at the opposite voltage of reference.
    m_reduced : float, tuple
        Ion chamber reading(s) at the reduced voltage.
    mu : int
        The MU delivered to measure the reference reading. E.g. 200.
    fff : bool
        Whether the beam is FFF or flat.
    tissue_correction : float
        Correction value to calibration to, e.g., muscle. A value of 1.0 means no correction (i.e. water).
    """

    @argue.options(chamber=KQ_PHOTONS.keys(), lead_foil=LEAD_OPTIONS.values())
    def __init__(
        self,
        *,
        institution: str = "",
        physicist: str = "",
        unit: str,
        measurement_date: str = "",
        temp: float,
        press: float,
        chamber: str,
        n_dw: float,
        p_elec: float,
        electrometer: str = "",
        measured_pdd10: Optional[float] = None,
        lead_foil: Optional[str] = None,
        clinical_pdd10: float,
        energy: int,
        fff: bool = False,
        voltage_reference: int,
        voltage_reduced: int,
        m_reference: NumberOrArray,
        m_opposite: NumberOrArray,
        m_reduced: NumberOrArray,
        mu: int,
        tissue_correction: float = 1.0,
        m_reference_adjusted: Optional[NumberOrArray] = None,
    ):
        super().__init__(
            temp=temp,
            press=press,
            chamber=chamber,
            n_dw=n_dw,
            p_elec=p_elec,
            measured_pdd10=measured_pdd10,
            energy=energy,
            voltage_reference=voltage_reference,
            voltage_reduced=voltage_reduced,
            m_reference=m_reference,
            m_opposite=m_opposite,
            m_reduced=m_reduced,
            clinical_pdd10=clinical_pdd10,
            mu=mu,
            tissue_correction=tissue_correction,
            lead_foil=lead_foil,
            electrometer=electrometer,
            m_reference_adjusted=m_reference_adjusted,
            institution=institution,
            physicist=physicist,
            unit=unit,
            measurement_date=measurement_date,
            fff=fff,
        )
        # add check for tpr vs pdd

    @property
    def pddx(self) -> float:
        """The photon-only PDD(10) value."""
        return pddx(
            pdd=self.measured_pdd10, energy=self.energy, lead_foil=self.lead_foil
        )

    @property
    def kq(self) -> float:
        """The chamber-specific beam quality correction factor."""
        return kq_photon_pddx(chamber=self.chamber, pddx=self.pddx)

    @property
    def dose_mu_10(self) -> float:
        """cGy/MU at a depth of 10cm."""
        return self.tissue_correction * self.m_corrected * self.kq * self.n_dw / self.mu

    @property
    def dose_mu_dmax(self) -> float:
        """cGy/MU at a depth of dmax."""
        return self.dose_mu_10 / (self.clinical_pdd10 / 100)

    @property
    def dose_mu_10_adjusted(self) -> float:
        """The dose/mu at 10cm depth after adjustment."""
        return (
            self.tissue_correction
            * self.m_corrected_adjustment
            * self.kq
            * self.n_dw
            / self.mu
        )

    @property
    def dose_mu_dmax_adjusted(self) -> float:
        """The dose/mu at dmax depth after adjustment."""
        return self.dose_mu_10_adjusted / (self.clinical_pdd10 / 100)

    def publish_pdf(
        self,
        filename: str,
        notes: Optional[list] = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ):
        """Publish (print) a PDF containing the analysis and quantitative results.

        Parameters
        ----------
        filename : str, file-like object
            The file to write the results to.
        notes : str, list
            Any notes to be added to the report. If a string, adds everything as one line.
            If a list, must be a list of strings; each string item will be a new line.
        open_file : bool
            Whether to open the file after creation. Will use the default PDF program.
        metadata : dict
            Any data that should be appended to every page of the report. This differs from notes in that
            metadata is at the top of every page while notes is at the bottom of the report.
        """
        was_adjusted = "Yes" if self.output_was_adjusted else "No"
        title = [
            "TG-51 Photon Report",
            f"{self.unit} - {self.energy} MV{' FFF' if self.fff else ''}",
        ]

        canvas = PylinacCanvas(filename, page_title=title, metadata=metadata)
        text = [
            "Site Data:",
            f"Institution: {self.institution}",
            f"Performed by: {self.physicist}",
            f"Measurement Date: {self.measurement_date}",
            f'Date of Report: {datetime.now().strftime("%A, %B %d, %Y")}',
            f"Unit: {self.unit}",
            f"Energy: {self.energy} MV {'FFF' if self.fff else ''}",
            "",
            "Instrumentation:",
            f"Chamber: {self.chamber}",
            f"Chamber Calibration Factor Ndw (cGy/nC): {self.n_dw:2.3f}",
            f"Electrometer: {self.electrometer}",
            f"Pelec: {self.p_elec:2.3f}",
            f"MU: {self.mu}",
            "",
            "Beam Quality:",
            f"Lead foil: {'No' if self.lead_foil is None else self.lead_foil}",
            f"Measured PDD(10){'' if self.lead_foil is None else 'Pb'} {self.measured_pdd10:2.2f}",
            f"Calculated PDD(10)x: {self.pddx:2.2f}",
            f"Determined kQ: {self.kq:2.3f}",
            "",
            "Chamber Corrections/Measurements:",
            f"Temperature (\N{DEGREE SIGN}C): {self.temp:2.1f}",
            f"Pressure (kPa): {self.press:2.1f}",
            f"Mraw @ ({self.voltage_reference}V, Reference) (nC): {self.m_reference}",
            f"Mraw @ ({self.voltage_reduced}V, Reduced) (nC): {self.m_reduced}",
            f"Mraw @ ({-self.voltage_reference}V, Opposite) (nC): {self.m_opposite}",
            f"Ptp: {self.p_tp:2.3f}",
            f"Pion: {self.p_ion:2.3f}",
            f"Ppol: {self.p_pol:2.3f}",
            "",
            "Dose Determination:",
            f"Fully corrected M (nC): {self.m_corrected:2.3f}",
            f"Tissue correction (e.g. muscle): {self.tissue_correction:2.3f}",
            f"Dose/MU @ 10cm depth (cGy): {self.dose_mu_10:2.3f}",
            f"Clinical PDD (%): {self.clinical_pdd10:2.2f}",
            f"Dose/MU @ dmax (cGy): {self.dose_mu_dmax:2.3f}",
            "",
            f"Output Adjustment?: {was_adjusted}",
        ]
        if was_adjusted == "Yes":
            text.append(
                f"Adjusted Mraw @ reference voltage (nC): {self.m_reference_adjusted}"
            )
            text.append(
                f"Adjusted fully corrected M (nC): {self.m_corrected_adjustment:2.3f}"
            )
            text.append(
                f"Adjusted Dose/MU @ 10cm depth (cGy): {self.dose_mu_10_adjusted:2.3f}"
            )
            text.append(
                f"Adjusted Dose/MU @ dmax (cGy): {self.dose_mu_dmax_adjusted:2.3f}"
            )
        canvas.add_text(text=text, location=(2, 25.5), font_size=12)
        if notes is not None:
            canvas.add_text(text="Notes:", location=(12, 6.5), font_size=14)
            canvas.add_text(text=notes, location=(12, 6))

        canvas.finish()

        if open_file:
            webbrowser.open(filename)


class TG51ElectronLegacy(TG51Base):
    """Class for calculating absolute dose to water using a cylindrical chamber in an electron beam.

    Parameters
    ----------
    institution : str
        Institution name.
    physicist : str
        Physicist performing calibration.
    unit : str
        Unit name; e.g. TrueBeam1.
    measurement_date : str
        Date of measurement. E.g. 10/22/2018.
    temp : float (17-27)
        The temperature in degrees Celsius.
    press : float (91-111)
        The value of pressure in kPa. Can be converted from mmHg and mbar; see :func:`~pylinac.calibration.tg51.mmHg2kPa` and :func:`~pylinac.calibration.tg51.mbar2kPa`.
    chamber : str
        Chamber model; only for bookkeeping.
    n_dw : float
        NDW value in Gy/nC. Given by the calibration laboratory.
    k_ecal : float
        Kecal value which is chamber specific. This value is the major difference between the legacy class and modern class where no kecal is needed.
    p_elec : float
        Electrometer correction factor; given by the calibration laboratory.
    clinical_pdd : float
        The PDD used to correct the dose back to dref.
    voltage_reference : float
        Reference voltage; i.e. voltage when taking the calibration measurement.
    voltage_reduced : float
        Reduced voltage; usually half of the reference voltage.
    m_reference : float, tuple
        Ion chamber reading(s) at the reference voltage.
    m_opposite : float, tuple
        Ion chamber reading(s) at the opposite voltage of reference.
    m_reduced : float, tuple
        Ion chamber reading(s) at the reduced voltage.
    mu : int
        The MU delivered to measure the reference reading. E.g. 200.
    i_50 : float
        Depth of 50% ionization.
    tissue_correction : float
        Correction value to calibration to, e.g., muscle. A value of 1.0 means no correction (i.e. water).
    """

    def __init__(
        self,
        *,
        institution: str = "",
        physicist: str = "",
        unit: str = "",
        measurement_date: str = "",
        energy: int,
        temp: float,
        press: float,
        chamber: str,
        k_ecal: float,
        n_dw: float,
        electrometer: str = "",
        p_elec: float,
        clinical_pdd: float,
        voltage_reference: int,
        voltage_reduced: int,
        m_reference: NumberOrArray,
        m_opposite: NumberOrArray,
        m_reduced: NumberOrArray,
        m_gradient: NumberOrArray,
        cone: str,
        mu: int,
        i_50: float,
        tissue_correction: float = 1.0,
        m_reference_adjusted=None,
    ):
        super().__init__(
            temp=temp,
            press=press,
            chamber=chamber,
            n_dw=n_dw,
            p_elec=p_elec,
            voltage_reference=voltage_reference,
            voltage_reduced=voltage_reduced,
            m_reference=m_reference,
            m_opposite=m_opposite,
            m_reduced=m_reduced,
            clinical_pdd=clinical_pdd,
            mu=mu,
            i_50=i_50,
            tissue_correction=tissue_correction,
            institution=institution,
            physicist=physicist,
            unit=unit,
            measurement_date=measurement_date,
            electrometer=electrometer,
            m_reference_adjusted=m_reference_adjusted,
            cone=cone,
            energy=energy,
            k_ecal=k_ecal,
            m_gradient=m_gradient,
        )

    @property
    def r_50(self) -> float:
        """Depth of the 50% dose value."""
        return r_50(i_50=self.i_50)

    @property
    def dref(self) -> float:
        """Depth of the reference point."""
        return d_ref(i_50=self.i_50)

    @property
    def pq_gr(self):
        """Gradient factor"""
        return pq_gr(m_dref_plus=self.m_gradient, m_dref=self.m_reference)

    @property
    def kq(self) -> float:
        """The kQ value using classic TG-51"""
        return self.k_ecal * kp_r50(r_50=self.r_50)

    @property
    def dose_mu_dref(self) -> float:
        """cGy/MU at the depth of Dref."""
        return (
            self.tissue_correction
            * self.m_corrected
            * self.kq
            * self.n_dw
            * self.pq_gr
            / self.mu
        )

    @property
    def dose_mu_dmax(self) -> float:
        """cGy/MU at the depth of dmax."""
        return self.dose_mu_dref / (self.clinical_pdd / 100)

    @property
    def dose_mu_dref_adjusted(self) -> float:
        """cGy/MU at the depth of Dref."""
        return (
            self.tissue_correction
            * self.m_corrected_adjustment
            * self.kq
            * self.n_dw
            * self.pq_gr
            / self.mu
        )

    @property
    def dose_mu_dmax_adjusted(self) -> float:
        """cGy/MU at the depth of dmax."""
        return self.dose_mu_dref_adjusted / (self.clinical_pdd / 100)

    def publish_pdf(
        self,
        filename: str,
        notes: Optional[list] = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ):
        """Publish (print) a PDF containing the analysis and quantitative results.

        Parameters
        ----------
        filename : str, file-like object
            The file to write the results to.
        notes : str, list
            Any notes to be added to the report. If a string, adds everything as one line.
            If a list, must be a list of strings; each string item will be a new line.
        open_file : bool
            Whether to open the file after creation. Will use the default PDF program.
        metadata : dict
            Any data that should be appended to every page of the report. This differs from notes in that
            metadata is at the top of every page while notes is at the bottom of the report.
        """
        was_adjusted = "Yes" if self.output_was_adjusted else "No"
        title = ["TG-51 Electron Report (Legacy)", f"{self.unit} - {self.energy} MeV"]

        canvas = PylinacCanvas(filename, page_title=title, metadata=metadata)
        text = [
            "Site Data:",
            f"Institution: {self.institution}",
            f"Performed by: {self.physicist}",
            f"Measurement Date: {self.measurement_date}",
            f'Date of Report: {datetime.now().strftime("%A, %B %d, %Y")}',
            f"Unit: {self.unit}",
            f"Energy: {self.energy} MeV",
            f"Cone: {self.cone}",
            f"MU: {self.mu}",
            "",
            "Instrumentation:",
            f"Chamber chamber: {self.chamber}",
            f"Chamber Calibration Factor Ndw (cGy/nC): {self.n_dw:2.3f}",
            f"Electrometer: {self.electrometer}",
            f"Pelec: {self.p_elec:2.2f}",
            "",
            "Beam Quality:",
            f"I50 (cm): {self.i_50:2.2f}",
            f"R50 (cm): {self.r_50:2.2f}",
            f"Dref (cm): {self.dref:2.2f}",
            f"Kecal: {self.k_ecal:2.3f}",
            f"kQ: {self.kq:2.3f}",
            "",
            "Chamber Corrections/Measurements:",
            f"Temperature (\N{DEGREE SIGN}C): {self.temp:2.1f}",
            f"Pressure (kPa): {self.press:2.1f}",
            f"Mraw @ ({self.voltage_reference}V, Reference) (nC): {self.m_reference}",
            f"Mraw @ ({self.voltage_reduced}V, Reduced) (nC): {self.m_reduced}",
            f"Mraw @ ({-self.voltage_reference}V, Opposite) (nC): {self.m_opposite}",
            f"Ptp: {self.p_tp:2.3f}",
            f"Pion: {self.p_ion:2.3f}",
            f"Ppol: {self.p_pol:2.3f}",
            f"Mraw @ Dref + 0.5rcav (nC): {self.m_gradient}",
            "",
            "Dose Determination:",
            f"Fully corrected M (nC): {self.m_corrected:2.3f}",
            f"Tissue correction (e.g. muscle): {self.tissue_correction:2.3f}",
            f"Dose/MU @ Dref depth (cGy): {self.dose_mu_dref:2.3f}",
            f"Clinical PDD (%): {self.clinical_pdd:2.2f}",
            f"Dose/MU @ dmax (cGy): {self.dose_mu_dmax:2.3f}",
            "",
            f"Output Adjustment?: {was_adjusted}",
        ]
        if was_adjusted == "Yes":
            text.append(
                f"Adjusted Mraw @ reference voltage (nC): {self.m_reference_adjustment}"
            )
            text.append(
                f"Adjusted fully corrected M (nC): {self.m_corrected_adjustment:2.3f}"
            )
            text.append(
                f"Adjusted Dose/MU @ dref depth (cGy): {self.dose_mu_dref_adjusted:2.3f}"
            )
            text.append(
                f"Adjusted Dose/MU @ dmax (cGy): {self.dose_mu_dmax_adjusted:2.3f}"
            )
        canvas.add_text(text=text, location=(2, 25.5), font_size=11)
        if notes is not None:
            canvas.add_text(text="Notes:", location=(12, 6.5), font_size=14)
            canvas.add_text(text=notes, location=(12, 6))

        canvas.finish()

        if open_file:
            webbrowser.open(filename)


class TG51ElectronModern(TG51Base):
    """Class for calculating absolute dose to water using a cylindrical chamber in an electron beam.

    .. warning::
        This class uses the values of Muir & Rogers. These values are likely to be included in the new TG-51
        addendum, but are not official. The results can be up to 1% different. Physicists should use their own
        judgement when deciding which class to use. To use a manual kecal value, Pgradient and the classic TG-51 equations use
        the :class:`~pylinac.calibration.tg51.TG51ElectronLegacy` class.

    Parameters
    ----------
    institution : str
        Institution name.
    physicist : str
        Physicist performing calibration.
    unit : str
        Unit name; e.g. TrueBeam1.
    measurement_date : str
        Date of measurement. E.g. 10/22/2018.
    press : float
        The value of pressure in kPa. Can be converted from mmHg and mbar; see :func:`~pylinac.calibration.tg51.mmHg2kPa` and :func:`~pylinac.calibration.tg51.mbar2kPa`.
    temp : float
        The temperature in Celsius.
    voltage_reference : int
        The reference voltage; i.e. the voltage for the calibration reading (e.g. 300V).
    voltage_reduced : int
        The reduced voltage, usually a fraction of the reference voltage (e.g. 150V).
    m_reference : array, float
        The reading(s) of the chamber at reference voltage.
    m_reduced : array, float
        The reading(s) of the chamber at the reduced voltage.
    m_opposite : array, float
        The reading(s) of the chamber at the opposite voltage from reference. Sign of the reading does not matter.
    k_elec : float
        The electrometer correction value given by the calibration laboratory. jyh,lykllp;ljljuhyk nmdrzj
    chamber : str
        Ion chamber model.
    n_dw : float
        NDW value in Gy/nC
    p_elec : float
        Electrometer correction given by the calibration laboratory.
    clinical_pdd : float
        The PDD used to correct the dose back to dref.
    mu : int
        MU delivered.
    i_50 : float
        Depth of 50% ionization
    tissue_correction : float
        Correction value to calibration to, e.g., muscle. A value of 1.0 means no correction (i.e. water).
    """

    def __init__(
        self,
        *,
        institution: str = "",
        physicist: str = "",
        unit: str = "",
        measurement_date: str = "",
        energy: int,
        temp: float,
        press: float,
        chamber: str,
        n_dw: float,
        electrometer: str = "",
        p_elec: float,
        clinical_pdd: float,
        voltage_reference: int,
        voltage_reduced: int,
        m_reference: NumberOrArray,
        m_opposite: NumberOrArray,
        m_reduced: NumberOrArray,
        cone: str,
        mu: int,
        i_50: float,
        tissue_correction: float,
        m_reference_adjusted=None,
    ):
        super().__init__(
            temp=temp,
            press=press,
            chamber=chamber,
            n_dw=n_dw,
            p_elec=p_elec,
            voltage_reference=voltage_reference,
            voltage_reduced=voltage_reduced,
            m_reference=m_reference,
            m_opposite=m_opposite,
            m_reduced=m_reduced,
            clinical_pdd=clinical_pdd,
            mu=mu,
            i_50=i_50,
            tissue_correction=tissue_correction,
            institution=institution,
            physicist=physicist,
            unit=unit,
            measurement_date=measurement_date,
            electrometer=electrometer,
            m_reference_adjusted=m_reference_adjusted,
            cone=cone,
            energy=energy,
        )

    @property
    def r_50(self) -> float:
        """Depth of the 50% dose value."""
        return r_50(i_50=self.i_50)

    @property
    def dref(self) -> float:
        """Depth of the reference point."""
        return d_ref(i_50=self.i_50)

    @property
    def kq(self) -> float:
        """The kQ value using the updated Muir & Rogers values from their 2014 paper, equation 11, or classically
        if kecal is passed."""
        return kq_electron(chamber=self.chamber, r_50=self.r_50)

    @property
    def dose_mu_dref(self) -> float:
        """cGy/MU at the depth of Dref."""
        return self.tissue_correction * self.m_corrected * self.kq * self.n_dw / self.mu

    @property
    def dose_mu_dmax(self) -> float:
        """cGy/MU at the depth of dmax."""
        return self.dose_mu_dref / (self.clinical_pdd / 100)

    @property
    def dose_mu_dref_adjusted(self) -> float:
        """cGy/MU at the depth of Dref."""
        return (
            self.tissue_correction
            * self.m_corrected_adjusted
            * self.kq
            * self.n_dw
            / self.mu
        )

    @property
    def dose_mu_dmax_adjusted(self) -> float:
        """cGy/MU at the depth of dmax."""
        return self.dose_mu_dref_adjusted / (self.clinical_pdd / 100)

    def publish_pdf(
        self,
        filename: str,
        notes: Optional[list] = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ):
        """Publish (print) a PDF containing the analysis and quantitative results.

        Parameters
        ----------
        filename : str, file-like object
            The file to write the results to.
        notes : str, list
            Any notes to be added to the report. If a string, adds everything as one line.
            If a list, must be a list of strings; each string item will be a new line.
        open_file : bool
            Whether to open the file after creation. Will use the default PDF program.
        metadata : dict
            Any data that should be appended to every page of the report. This differs from notes in that
            metadata is at the top of every page while notes is at the bottom of the report.
        """
        was_adjusted = "Yes" if self.output_was_adjusted else "No"
        title = ["TG-51 Electron Report (Modern)", f"{self.unit} - {self.energy} MeV"]

        canvas = PylinacCanvas(filename, page_title=title, metadata=metadata)
        text = [
            "Site Data:",
            f"Institution: {self.institution}",
            f"Performed by: {self.physicist}",
            f"Measurement Date: {self.measurement_date}",
            f'Date of Report: {datetime.now().strftime("%A, %B %d, %Y")}',
            f"Unit: {self.unit}",
            f"Energy: {self.energy} MeV",
            f"Cone: {self.cone}",
            f"MU: {self.mu}",
            "",
            "Instrumentation:",
            f"Chamber: {self.chamber}",
            f"Chamber Calibration Factor Ndw (cGy/nC): {self.n_dw:2.3f}",
            f"Electrometer: {self.electrometer}",
            f"Pelec: {self.p_elec:2.2f}",
            "",
            "Beam Quality:",
            f"I50 (cm): {self.i_50:2.2f}",
            f"R50 (cm): {self.r_50:2.2f}",
            f"Dref (cm): {self.dref:2.2f}",
            f"Calculated kQ: {self.kq:2.3f}",
            "",
            "Chamber Corrections/Measurements:",
            f"Temperature (\N{DEGREE SIGN}C): {self.temp:2.1f}",
            f"Pressure (kPa): {self.press:2.1f}",
            f"Mraw @ ({self.voltage_reference}V, Reference) (nC): {self.m_reference}",
            f"Mraw @ ({self.voltage_reduced}V, Reduced) (nC): {self.m_reduced}",
            f"Mraw @ ({-self.voltage_reference}V, Opposite) (nC): {self.m_opposite}",
            f"Ptp: {self.p_tp:2.3f}",
            f"Pion: {self.p_ion:2.3f}",
            f"Ppol: {self.p_pol:2.3f}",
            "",
            "Dose Determination:",
            f"Fully corrected M (nC): {self.m_corrected:2.3f}",
            f"Tissue correction (e.g. muscle): {self.tissue_correction:2.3f}",
            f"Dose/MU @ Dref depth (cGy): {self.dose_mu_dref:2.3f}",
            f"Clinical PDD (%): {self.clinical_pdd:2.2f}",
            f"Dose/MU @ dmax (cGy): {self.dose_mu_dmax:2.3f}",
            "",
            f"Output Adjustment?: {was_adjusted}",
        ]
        if was_adjusted == "Yes":
            text.append(
                f"Adjusted corrected M @ reference voltage (nC): {self.m_corrected_adjustment}"
            )
            text.append(
                f"Adjusted fully corrected M (nC): {self.m_corrected_adjustment:2.3f}"
            )
            text.append(
                f"Adjusted Dose/MU @ dref depth (cGy): {self.dose_mu_dref_adjusted:2.3f}"
            )
            text.append(
                f"Adjusted Dose/MU @ dmax (cGy): {self.dose_mu_dmax_adjusted:2.3f}"
            )
        canvas.add_text(text=text, location=(2, 25.5), font_size=11)
        if notes is not None:
            canvas.add_text(text="Notes:", location=(12, 6.5), font_size=14)
            canvas.add_text(text=notes, location=(12, 6))

        canvas.finish()

        if open_file:
            webbrowser.open(filename)



import dataclasses
import io
import math
import textwrap
import webbrowser
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import Union, List, Dict, Optional, Tuple

import numpy as np
from matplotlib import pyplot as plt
from scipy import ndimage

from .core import pdf
from .core.geometry import Line, Point
from .core.mtf import MTF
from .core.profile import SingleProfile, Interpolation
from .core.roi import HighContrastDiskROI, RectangleROI
from .core.utilities import ResultBase
from .ct import CatPhanBase, CatPhanModule, ThicknessROI, Slice, get_regions

# CT
CT_UNIFORMITY_MODULE_OFFSET_MM = 70
CT_SPATIAL_RESOLUTION_MODULE_OFFSET_MM = 100
CT_LOW_CONTRAST_MODULE_OFFSET_MM = 30

# MR
MR_SLICE11_MODULE_OFFSET_MM = 100
MR_GEOMETRIC_DISTORTION_MODULE_OFFSET_MM = 40
MR_UNIFORMITY_MODULE_OFFSET_MM = 60


class CTModule(CatPhanModule):
    attr_name = "ct_calibration_module"
    roi_dist_mm = 63
    roi_radius_mm = 10
    roi_settings = {
        "Air": {"angle": 45, "distance": roi_dist_mm, "radius": roi_radius_mm},
        "Poly": {"angle": 225, "distance": roi_dist_mm, "radius": roi_radius_mm},
        "Acrylic": {"angle": 135, "distance": roi_dist_mm, "radius": roi_radius_mm},
        "Bone": {"angle": -45, "distance": roi_dist_mm, "radius": roi_radius_mm},
        "Water": {"angle": 180, "distance": roi_dist_mm, "radius": roi_radius_mm},
    }
    window_min = -200
    window_max = 200


@dataclass
class CTModuleOutput:
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    offset: int
    roi_distance_from_center_mm: int
    roi_radius_mm: int
    roi_settings: dict
    rois: dict


class UniformityModule(CatPhanModule):
    """Class for analysis of the Uniformity slice of the CTP module. Measures 5 ROIs around the slice that
    should all be close to the same value.
    """

    attr_name = "uniformity_module"
    common_name = "HU Uniformity"
    roi_dist_mm = 66
    roi_radius_mm = 11
    roi_settings = {
        "Top": {"angle": -90, "distance": roi_dist_mm, "radius": roi_radius_mm},
        "Right": {"angle": 0, "distance": roi_dist_mm, "radius": roi_radius_mm},
        "Bottom": {"angle": 90, "distance": roi_dist_mm, "radius": roi_radius_mm},
        "Left": {"angle": 180, "distance": roi_dist_mm, "radius": roi_radius_mm},
        "Center": {"angle": 0, "distance": 0, "radius": roi_radius_mm},
    }
    window_min = -50
    window_max = 50


@dataclass
class UniformityModuleOutput(CTModuleOutput):
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    center_roi_stdev: float


class SpatialResolutionModule(CatPhanModule):
    """Class for analysis of the Uniformity slice of the CTP module. Measures 5 ROIs around the slice that
    should all be close to the same value.
    """

    attr_name = "spatial_resolution_module"
    common_name = "Spatial Resolution"
    rois: Dict[str, HighContrastDiskROI]
    roi_dist_mm = 70
    roi_radius_mm = 6
    roi_settings = {
        "10oclock": {
            "angle": -135,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
            "lp/mm": 0.4,
        },
        "9oclock": {
            "angle": -180,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
            "lp/mm": 0.5,
        },
        "7oclock": {
            "angle": 135,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
            "lp/mm": 0.6,
        },
        "6oclock": {
            "angle": 90,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
            "lp/mm": 0.7,
        },
        "4oclock": {
            "angle": 45,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
            "lp/mm": 0.8,
        },
        "3oclock": {
            "angle": 0,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
            "lp/mm": 0.9,
        },
        "2oclock": {
            "angle": -45,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
            "lp/mm": 1.0,
        },
        "12oclock": {
            "angle": -90,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
            "lp/mm": 1.2,
        },
    }

    def _setup_rois(self) -> None:
        for name, setting in self.roi_settings.items():
            self.rois[name] = HighContrastDiskROI(
                self.image,
                setting["angle_corrected"],
                setting["radius_pixels"],
                setting["distance_pixels"],
                self.phan_center,
                contrast_threshold=1.0,  # fixed to 1 so everything passes. We aren't evaluating pass/fail here
            )

    @property
    def mtf(self) -> MTF:
        spacings = [roi["lp/mm"] for roi in self.roi_settings.values()]
        return MTF.from_high_contrast_diskset(
            spacings=spacings, diskset=list(self.rois.values())
        )

    def plot_rois(self, axis: plt.Axes) -> None:
        """Plot the ROIs to the axis. Override to set the color"""
        for roi, mtf in zip(self.rois.values(), self.mtf.norm_mtfs.values()):
            roi.plot2axes(axis, edgecolor="g")


@dataclass
class SpatialResolutionModuleOutput(CTModuleOutput):
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    lpmm_to_rmtf: dict


class LowContrastModule(CatPhanModule):
    """Class for analysis of the Uniformity slice of the CTP module. Measures 5 ROIs around the slice that
    should all be close to the same value.
    """

    attr_name = "low_contrast_module"
    common_name = "Low Contrast"
    roi_dist_mm = 60
    roi_radius_mm = 6
    nominal_value = 0
    roi_settings = {
        "ROI": {"angle": -90, "distance": roi_dist_mm, "radius": roi_radius_mm},
    }
    background_roi_settings = {
        "ROI": {"angle": -115, "distance": roi_dist_mm, "radius": roi_radius_mm},
    }
    window_min = 50
    window_max = 150

    def cnr(self) -> float:
        """Given in the guidance doc as |A-B|/SD where A is the contrast ROI, B is the background, and SD is stdev of B"""
        return (
            abs(self.rois["ROI"].pixel_value - self.background_rois["ROI"].pixel_value)
            / self.background_rois["ROI"].std
        )


@dataclass
class LowContrastModuleOutput(CTModuleOutput):
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    cnr: float


@dataclass
class ACRCTResult(ResultBase):
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    phantom_model: str  #:
    phantom_roll_deg: float  #:
    origin_slice: int  #:
    num_images: int  #:
    ct_module: CTModuleOutput  #:
    uniformity_module: UniformityModuleOutput  #:
    low_contrast_module: LowContrastModuleOutput  #:
    spatial_resolution_module: SpatialResolutionModuleOutput  #:


class ACRCT(CatPhanBase):
    _model = "ACR CT 464"
    catphan_radius_mm = 100
    air_bubble_radius_mm = 18
    min_num_images = 4
    localization_radius = 70
    results: dict
    ct_calibration_module = CTModule
    low_contrast_module = LowContrastModule
    spatial_resolution_module = SpatialResolutionModule
    uniformity_module = UniformityModule
    clear_borders = False

    def plot_analyzed_subimage(self, *args, **kwargs):
        raise NotImplementedError("Use `plot_images`")

    def save_analyzed_subimage(self, *args, **kwargs):
        raise NotImplementedError("Use `save_images`")

    def analyze(self) -> None:
        """Analyze the ACR CT phantom"""
        self.ct_calibration_module = self.ct_calibration_module(
            self, offset=0, clear_borders=self.clear_borders
        )
        self.uniformity_module = self.uniformity_module(
            self,
            offset=CT_UNIFORMITY_MODULE_OFFSET_MM,
            clear_borders=self.clear_borders,
        )
        self.spatial_resolution_module = self.spatial_resolution_module(
            self,
            offset=CT_SPATIAL_RESOLUTION_MODULE_OFFSET_MM,
            clear_borders=self.clear_borders,
        )
        self.low_contrast_module = self.low_contrast_module(
            self,
            offset=CT_LOW_CONTRAST_MODULE_OFFSET_MM,
            clear_borders=self.clear_borders,
        )

    def plot_analyzed_image(self, show: bool = True, **plt_kwargs) -> plt.Figure:
        """Plot the analyzed image

        Parameters
        ----------
        show
            Whether to show the image.
        plt_kwargs
            Keywords to pass to matplotlib for figure customization.
        """
        # set up grid and axes
        fig = plt.figure(**plt_kwargs)
        grid_size = (2, 3)
        hu_ax = plt.subplot2grid(grid_size, (0, 0))
        self.ct_calibration_module.plot(hu_ax)
        unif_ax = plt.subplot2grid(grid_size, (0, 1))
        self.uniformity_module.plot(unif_ax)
        sr_ax = plt.subplot2grid(grid_size, (0, 2))
        self.spatial_resolution_module.plot(sr_ax)
        locon_ax = plt.subplot2grid(grid_size, (1, 0))
        self.low_contrast_module.plot(locon_ax)
        spatial_res_graph = plt.subplot2grid(grid_size, (1, 1), colspan=2)
        self.spatial_resolution_module.mtf.plot(spatial_res_graph)

        # finish up
        plt.tight_layout()
        if show:
            plt.show()
        return fig

    def save_analyzed_image(
        self, filename: Union[str, Path, BytesIO], **plt_kwargs
    ) -> None:
        """Save the analyzed image to disk or stream

        Parameters
        ----------
        filename
            Where to save the image to
        plt_kwargs
            Keywords to pass to matplotlib for figure customization.
        """
        fig = self.plot_analyzed_image(show=False, **plt_kwargs)
        fig.savefig(filename)

    def plot_images(self, show: bool = True, **plt_kwargs) -> Dict[str, plt.Figure]:
        """Plot all the individual images separately

        Parameters
        ----------
        show
            Whether to show the images.
        plt_kwargs
            Keywords to pass to matplotlib for figure customization.
        """
        figs = {}
        # plot the images
        modules = {
            "hu": self.ct_calibration_module,
            "uniformity": self.uniformity_module,
            "spatial resolution": self.spatial_resolution_module,
            "low contrast": self.low_contrast_module,
        }
        for key, module in modules.items():
            fig, ax = plt.subplots(**plt_kwargs)
            module.plot(ax)
            figs[key] = fig
        # plot the one-off MTF image
        fig, ax = plt.subplots(**plt_kwargs)
        figs["mtf"] = fig
        self.spatial_resolution_module.mtf.plot(ax)
        plt.tight_layout()

        if show:
            plt.show()
        return figs

    def save_images(
        self,
        directory: Optional[Union[Path, str]] = None,
        to_stream: bool = False,
        **plt_kwargs,
    ) -> List[Union[Path, BytesIO]]:
        """Save separate images to disk or stream.

        Parameters
        ----------
        directory
            The directory to write the images to. If None, will use current working directory
        to_stream
            Whether to write to stream or disk. If True, will return streams. Directory is ignored in that scenario.
        plt_kwargs
            Keywords to pass to matplotlib for figure customization.
        """
        figs = self.plot_images(show=False, **plt_kwargs)
        paths = []
        for name, fig in figs.items():
            if to_stream:
                path = io.BytesIO()
            else:
                destination = Path(directory) or Path.cwd()
                path = (destination / name).with_suffix(".png").absolute()
            fig.savefig(path)
            paths.append(path)
        return paths

    def find_phantom_roll(self, func=lambda roi: roi.bbox_area) -> float:
        """Determine the "roll" of the phantom.

        Only difference of base method is that we sort the ROIs by size,
        not by being in the center since the two we're looking for are both right-sided.
        """
        return super().find_phantom_roll(func)

    def results(self) -> str:
        """Return the results of the analysis as a string. Use with print()."""
        string = (
            f"\n - ACR CT 464 QA Test - \n"
            f"HU ROIs: {self.ct_calibration_module.roi_vals_as_str}\n"
            f"Contrast to Noise Ratio: {self.low_contrast_module.cnr():2.2f}\n"
            f"Uniformity ROIs: {self.uniformity_module.roi_vals_as_str}\n"
            f'Uniformity Center ROI standard deviation: {self.uniformity_module.rois["Center"].std:2.2f}\n'
            f"MTF 50% (lp/mm): {self.spatial_resolution_module.mtf.relative_resolution(50):2.2f}\n"
        )
        return string

    def results_data(self, as_dict=False) -> Union[ACRCTResult, dict]:
        data = ACRCTResult(
            phantom_model="ACR CT 464",
            phantom_roll_deg=self.catphan_roll,
            origin_slice=self.origin_slice,
            num_images=self.num_images,
            ct_module=CTModuleOutput(
                offset=0,
                roi_distance_from_center_mm=self.ct_calibration_module.roi_dist_mm,
                roi_radius_mm=self.ct_calibration_module.roi_radius_mm,
                roi_settings=self.ct_calibration_module.roi_settings,
                rois={
                    name: roi.pixel_value
                    for name, roi in self.ct_calibration_module.rois.items()
                },
            ),
            uniformity_module=UniformityModuleOutput(
                offset=CT_UNIFORMITY_MODULE_OFFSET_MM,
                roi_distance_from_center_mm=self.uniformity_module.roi_dist_mm,
                roi_radius_mm=self.uniformity_module.roi_radius_mm,
                roi_settings=self.uniformity_module.roi_settings,
                rois={
                    name: roi.pixel_value
                    for name, roi in self.uniformity_module.rois.items()
                },
                center_roi_stdev=self.uniformity_module.rois["Center"].std,
            ),
            spatial_resolution_module=SpatialResolutionModuleOutput(
                offset=CT_SPATIAL_RESOLUTION_MODULE_OFFSET_MM,
                roi_distance_from_center_mm=self.spatial_resolution_module.roi_dist_mm,
                roi_radius_mm=self.spatial_resolution_module.roi_radius_mm,
                roi_settings=self.spatial_resolution_module.roi_settings,
                rois={
                    name: roi.pixel_value
                    for name, roi in self.spatial_resolution_module.rois.items()
                },
                lpmm_to_rmtf=self.spatial_resolution_module.mtf.norm_mtfs,
            ),
            low_contrast_module=LowContrastModuleOutput(
                offset=CT_LOW_CONTRAST_MODULE_OFFSET_MM,
                roi_distance_from_center_mm=self.low_contrast_module.roi_dist_mm,
                roi_radius_mm=self.low_contrast_module.roi_radius_mm,
                roi_settings=self.low_contrast_module.roi_settings,
                rois={
                    name: roi.pixel_value
                    for name, roi in self.low_contrast_module.rois.items()
                },
                cnr=self.low_contrast_module.cnr(),
            ),
        )
        if as_dict:
            return dataclasses.asdict(data)
        return data

    def publish_pdf(
        self,
        filename: Union[str, Path],
        notes: Optional[str] = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ) -> None:
        """Publish (print) a PDF containing the analysis and quantitative results.

        Parameters
        ----------
        filename : (str, file-like object}
            The file to write the results to.
        notes : str, list of strings
            Text; if str, prints single line.
            If list of strings, each list item is printed on its own line.
        open_file : bool
            Whether to open the file using the default program after creation.
        metadata : dict
            Extra data to be passed and shown in the PDF. The key and value will be shown with a colon.
            E.g. passing {'Author': 'James', 'Unit': 'TrueBeam'} would result in text in the PDF like:
            --------------
            Author: James
            Unit: TrueBeam
            --------------
        """
        analysis_title = f"{self._model} Analysis"
        texts = [
            " - ACR CT 464 Results - ",
            f"HU Linearity ROIs: {self.ct_calibration_module.roi_vals_as_str}",
            f"Low contrast visibility: {self.low_contrast_module.cnr():2.2f}",
            f"Uniformity ROIs: {self.uniformity_module.roi_vals_as_str}",
        ]
        analysis_images = self.save_images(to_stream=True)

        canvas = pdf.PylinacCanvas(
            filename, page_title=analysis_title, metadata=metadata
        )
        if notes is not None:
            canvas.add_text(text="Notes:", location=(1, 4.5), font_size=14)
            canvas.add_text(text=notes, location=(1, 4))

        for idx, text in enumerate(texts):
            canvas.add_text(text=text, location=(1.5, 23 - idx * 0.5))
        for page, img in enumerate(analysis_images):
            canvas.add_new_page()
            canvas.add_image(img, location=(1, 5), dimensions=(18, 18))
        canvas.finish()

        if open_file:
            webbrowser.open(filename)


class MRSlice11PositionModule(CatPhanModule):
    common_name = "Slice Position, Slice 11"
    roi_settings = {
        "Left": {"width": 2, "height": 25, "distance": 65, "angle": 2.5},
        "Right": {"width": 2, "height": 25, "distance": 65, "angle": -2.5},
    }
    rois: dict = {}

    def _setup_rois(self) -> None:
        for name, setting in self.roi_settings.items():
            # angle is +90 because pointing right is 0, and these rois move downward, not rightward
            self.rois[name] = ThicknessROI(
                self.image,
                setting["width_pixels"],
                setting["height_pixels"],
                self.catphan_roll - 90 + setting["angle"],
                setting["distance_pixels"],
                self.phan_center,
            )

    @property
    def bar_difference_mm(self) -> float:
        """The difference in height between the two angled bars"""
        left_array = self.rois["Left"].long_profile.values
        left_mid_height = (left_array.max() - left_array.min()) / 2 + left_array.min()
        left_idx = find_nearest_idx(left_array, left_mid_height)
        right_array = self.rois["Right"].long_profile.values
        right_mid_height = (
            right_array.max() - right_array.min()
        ) / 2 + right_array.min()
        right_idx = find_nearest_idx(right_array, right_mid_height)
        return (right_idx - left_idx) * self.mm_per_pixel

    @property
    def slice_shift_mm(self) -> float:
        """The effective shift in phantom position in the S/I direction. Because bars are at 45 degrees, the shift is half the bar difference"""
        return self.bar_difference_mm / 2


@dataclass
class MRSlice11ModuleOutput:
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    offset: int
    roi_settings: dict
    rois: dict
    bar_difference_mm: float
    slice_shift_mm: float


class MRSlice1Module(CatPhanModule):
    common_name = "Slice 1 (Thickness, Offset, Resolution)"
    slice_lines: Dict[str, Line]
    thickness_rois: Dict[str, ThicknessROI] = {}
    thickness_roi_settings = {
        "Top": {"width": 100, "height": 3.5, "distance": -3},
        "Bottom": {"width": 100, "height": 3.5, "distance": 2.5},
    }
    roi_settings = {
        "Row Reference": {"radius": 9, "distance": 58, "angle": 135, "lp/mm": 0},
        "Col Reference": {"radius": 9, "distance": 58, "angle": 135, "lp/mm": 0},
        "Row 1.1": {"radius": 3, "distance": 41, "angle": 115, "lp/mm": 1 / 1.1},
        "Col 1.1": {"radius": 3, "distance": 45, "angle": 103, "lp/mm": 1 / 1.1},
        "Row 1.0": {"radius": 3, "distance": 37.5, "angle": 81, "lp/mm": 1.0},
        "Col 1.0": {"radius": 3, "distance": 45, "angle": 74, "lp/mm": 1.0},
        "Row 0.9": {"radius": 2, "distance": 47.5, "angle": 52, "lp/mm": 1 / 0.9},
        "Col 0.9": {"radius": 2, "distance": 56, "angle": 51, "lp/mm": 1 / 0.9},
    }
    position_roi_settings = {
        "Left": {"width": 2, "height": 25, "distance": 65, "angle": 2.5},
        "Right": {"width": 2, "height": 25, "distance": 65, "angle": -2.5},
    }
    position_rois: dict = {}
    rois: Dict[str, HighContrastDiskROI]
    spacings = [0, 1 / 1.1, 1, 1 / 0.9]

    def _setup_rois(self) -> None:
        # thickness
        for name, setting in self.thickness_roi_settings.items():
            # angle is +90 because pointing right is 0, and these rois move downward, not rightward
            self.thickness_rois[name] = ThicknessROI(
                self.image,
                setting["width_pixels"],
                setting["height_pixels"],
                self.catphan_roll + 90,
                setting["distance_pixels"],
                self.phan_center,
            )
        # spatial res
        for name, setting in self.roi_settings.items():
            self.rois[name] = HighContrastDiskROI(
                self.image,
                setting["angle_corrected"],
                setting["radius_pixels"],
                setting["distance_pixels"],
                self.phan_center,
                contrast_threshold=1.0,  # fixed to 1 so everything passes. We aren't evaluating pass/fail here
            )
        # slice position
        for name, setting in self.position_roi_settings.items():
            # angle is +90 because pointing right is 0, and these rois move downward, not rightward
            self.position_rois[name] = ThicknessROI(
                self.image,
                setting["width_pixels"],
                setting["height_pixels"],
                self.catphan_roll - 90 + setting["angle"],
                setting["distance_pixels"],
                self.phan_center,
            )

    def plot_rois(self, axis: plt.Axes) -> None:
        for roi in self.position_rois.values():
            roi.plot2axes(axis, edgecolor="blue")
        for roi in self.thickness_rois.values():
            roi.plot2axes(axis, edgecolor="blue")
        for roi, mtf in zip(self.rois.values(), self.rois.values()):
            roi.plot2axes(axis, edgecolor="g")

    @property
    def bar_difference_mm(self) -> float:
        """The difference in height between the two angled bars"""
        left_array = self.position_rois["Left"].long_profile.values
        left_mid_height = (left_array.max() - left_array.min()) / 2 + left_array.min()
        left_idx = find_nearest_idx(left_array, left_mid_height)
        right_array = self.position_rois["Right"].long_profile.values
        right_mid_height = (
            right_array.max() - right_array.min()
        ) / 2 + right_array.min()
        right_idx = find_nearest_idx(right_array, right_mid_height)
        return (right_idx - left_idx) * self.mm_per_pixel

    @property
    def slice_shift_mm(self) -> float:
        """The effective shift in phantom position in the S/I direction. Because bars are at 45 degrees, the shift is half the bar difference"""
        return self.bar_difference_mm / 2

    @property
    def measured_slice_thickness_mm(self) -> float:
        """The slice thickness as determined by the two angled ROIs in the center of Slice 1"""
        top = self.thickness_rois["Top"].wire_fwhm * self.mm_per_pixel
        bottom = self.thickness_rois["Bottom"].wire_fwhm * self.mm_per_pixel
        return 0.2 * (top * bottom) / (top + bottom)

    @property
    def row_mtf(self) -> MTF:
        """The MTF of the spatial resolution module looking at the row-wise ROIs"""
        return MTF.from_high_contrast_diskset(
            spacings=self.spacings,
            diskset=list(roi for name, roi in self.rois.items() if "Row" in name),
        )

    @property
    def col_mtf(self) -> MTF:
        """The MTF of the spatial resolution module looking at the column-wise ROIs"""
        return MTF.from_high_contrast_diskset(
            spacings=self.spacings,
            diskset=list(roi for name, roi in self.rois.items() if "Col" in name),
        )


@dataclass
class MRSlice1ModuleOutput:
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    offset: int
    roi_settings: dict
    rois: dict
    bar_difference_mm: float
    slice_shift_mm: float
    measured_slice_thickness_mm: float
    row_mtf_50: float
    col_mtf_50: float


class MRUniformityModule(CatPhanModule):
    """Class for analysis of the Uniformity slice of the CTP module. Measures 5 ROIs around the slice that
    should all be close to the same value.
    """

    common_name = "Signal Uniformity"
    roi_settings = {
        "Center": {
            "angle": 90,
            "distance": 5,
            "radius": 80,
        },  # 80 radius ~= 200cm2, per the manual
    }
    ghost_roi_settings = {
        # size of ~900mm2 per the manual
        "Top": {"angle": -90, "distance": 110, "width": 60, "height": 15},
        "Bottom": {"angle": 90, "distance": 110, "width": 60, "height": 15},
        "Left": {"angle": 180, "distance": 110, "width": 15, "height": 60},
        "Right": {"angle": 0, "distance": 110, "width": 15, "height": 60},
    }
    ghost_rois: dict = {}

    def __init__(self, catphan, offset):
        self.tesla = float(catphan.dicom_stack.metadata.MagneticFieldStrength)
        super().__init__(catphan, tolerance=None, offset=offset)

    def _setup_rois(self) -> None:
        super()._setup_rois()
        for name, roi in self.ghost_roi_settings.items():
            self.ghost_rois[name] = RectangleROI(
                self.image,
                roi["width_pixels"],
                roi["height_pixels"],
                roi["angle"] + self.catphan_roll,
                roi["distance_pixels"],
                self.phan_center,
            )

    def plot_rois(self, axis: plt.Axes) -> None:
        super().plot_rois(axis)
        for roi in self.ghost_rois.values():
            roi.plot2axes(axis, edgecolor="yellow")

    @property
    def percent_image_uniformity(self) -> float:
        """PIU value calculated via section 5.3 of the manual"""
        piu_high = np.percentile(self.rois["Center"].pixel_values, 99)
        piu_low = np.percentile(self.rois["Center"].pixel_values, 1)
        return 100 * (1 - ((piu_high - piu_low) / (piu_high + piu_low)))

    @property
    def piu_passed(self) -> bool:
        """Section 5.4"""
        if self.tesla < 3:
            return self.percent_image_uniformity > 85
        else:
            return self.percent_image_uniformity > 80

    @property
    def ghosting_ratio(self) -> float:
        """Ghosting ratio of section 6.3 of the manual"""
        top = self.ghost_rois["Top"].pixel_value
        bottom = self.ghost_rois["Bottom"].pixel_value
        left = self.ghost_rois["Left"].pixel_value
        right = self.ghost_rois["Right"].pixel_value
        return abs(
            ((top + bottom) - (left + right)) / (2 * self.rois["Center"].pixel_value)
        )

    @property
    def psg(self) -> float:
        """Percent Signal Ghosting"""
        return self.ghosting_ratio * 100

    @property
    def psg_passed(self) -> bool:
        """Whether the PSG is within tolerance"""
        return self.psg < 3.0


@dataclass
class MRUniformityModuleOutput:
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    offset: int
    roi_settings: dict
    rois: dict
    ghost_roi_settings: dict
    ghose_rois: dict
    psg: float
    ghosting_ratio: float
    piu_passed: bool
    piu: float


class GeometricDistortionModule(CatPhanModule):
    """Class for analysis of the Uniformity slice of the CTP module. Measures 5 ROIs around the slice that
    should all be close to the same value.
    """

    common_name = "Geometric Distortion"
    profiles: dict

    def _setup_rois(self) -> None:
        self.profiles = {}
        bin_image = self.image.as_binary(threshold=np.percentile(self.image, 60))
        bin_image = ndimage.binary_fill_holes(bin_image).astype(float)
        # calculate horizontal
        data = bin_image[int(self.phan_center.y), :]
        prof = SingleProfile(
            data, interpolation=Interpolation.NONE, dpmm=1 / self.mm_per_pixel
        )
        fwhm = prof.fwxm_data()
        line = Line(
            Point(fwhm["left index (rounded)"], self.phan_center.y),
            Point(fwhm["right index (rounded)"], self.phan_center.y),
        )
        self.profiles["horizontal"] = {
            "width (mm)": fwhm["width (exact) mm"],
            "line": line,
        }
        # calculate vertical
        data = bin_image[:, int(self.phan_center.x)]
        prof = SingleProfile(
            data, interpolation=Interpolation.NONE, dpmm=1 / self.mm_per_pixel
        )
        fwhm = prof.fwxm_data()
        line = Line(
            Point(self.phan_center.x, fwhm["left index (rounded)"]),
            Point(self.phan_center.x, fwhm["right index (rounded)"]),
        )
        self.profiles["vertical"] = {
            "width (mm)": fwhm["width (exact) mm"],
            "line": line,
        }
        # calculate negative diagonal
        # calculate slope equation intercept
        # b = y - (+1)x
        b = self.phan_center.y - self.phan_center.x
        xs = np.arange(0, self.image.shape[1])
        ys = xs + b
        coords = ndimage.map_coordinates(bin_image, [ys, xs], order=1, mode="mirror")
        # pixels are now diagonal and thus spacing between pixels is now the hypotenuse
        prof = SingleProfile(
            coords,
            interpolation=Interpolation.NONE,
            dpmm=1 / (self.mm_per_pixel * math.sqrt(2)),
        )
        fwhm = prof.fwxm_data()
        line = Line(
            Point(xs[fwhm["left index (rounded)"]], ys[fwhm["left index (rounded)"]]),
            Point(xs[fwhm["right index (rounded)"]], ys[fwhm["right index (rounded)"]]),
        )
        self.profiles["negative diagonal"] = {
            "width (mm)": fwhm["width (exact) mm"],
            "line": line,
        }
        # calculate positive diagonal
        # calculate slope equation intercept
        # b = y - (-1)x
        b = self.phan_center.y + self.phan_center.x
        ys = -xs + b
        coords = ndimage.map_coordinates(bin_image, [ys, xs], order=1, mode="mirror")
        prof = SingleProfile(
            coords,
            interpolation=Interpolation.NONE,
            dpmm=1 / (self.mm_per_pixel * math.sqrt(2)),
        )
        fwhm = prof.fwxm_data()
        line = Line(
            Point(xs[fwhm["left index (rounded)"]], ys[fwhm["left index (rounded)"]]),
            Point(xs[fwhm["right index (rounded)"]], ys[fwhm["right index (rounded)"]]),
        )
        self.profiles["positive diagonal"] = {
            "width (mm)": fwhm["width (exact) mm"],
            "line": line,
        }

    def plot_rois(self, axis: plt.Axes):
        for name, profile_data in self.profiles.items():
            profile_data["line"].plot2axes(axis, width=2, color="blue")

    def distances(self) -> dict:
        """The measurements of the phantom size for all 4 lines in mm"""
        return {name: f"{p['width (mm)']:2.2f}mm" for name, p in self.profiles.items()}


@dataclass
class MRGeometricDistortionModuleOutput:
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    offset: int
    profiles: dict
    distances: dict


@dataclass
class ACRMRIResult(ResultBase):
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    phantom_model: str  #:
    phantom_roll_deg: float  #:
    origin_slice: int  #:
    num_images: int  #:
    slice1: MRSlice1ModuleOutput  #:
    slice11: MRSlice11ModuleOutput  #:
    uniformity_module: MRUniformityModuleOutput  #:
    geometric_distortion_module: MRGeometricDistortionModuleOutput  #:


class ACRMRILarge(CatPhanBase):
    _model = "ACR MRI Large"
    catphan_radius_mm = 100
    min_num_images = 4
    air_bubble_radius_mm = 20

    def plot_analyzed_subimage(self, *args, **kwargs):
        raise NotImplementedError("Use `plot_images`")

    def save_analyzed_subimage(self, *args, **kwargs):
        raise NotImplementedError("Use `save_images`")

    def localize(self) -> None:
        self.origin_slice = 1
        self.catphan_roll = self.find_phantom_roll()

    def find_phantom_roll(self) -> float:
        """Determine the "roll" of the phantom. This algorithm uses the circular left-upper hole on slice 1 as the reference

        Returns
        -------
        float : the angle of the phantom in **degrees**.
        """
        # get edges and make ROIs from it
        slice = Slice(self, self.origin_slice)
        larr, regions, _ = get_regions(slice)
        try:
            # find appropriate ROIs and grab the two most centrally positioned ones
            circle_bubbles = [
                r
                for r in regions
                if (self._is_right_area(r) and self._is_right_eccentricity(r))
            ]
            exact_size = np.pi * ((self.air_bubble_radius_mm / self.mm_per_pixel) ** 2)
            most_similar_bubble = sorted(
                circle_bubbles, key=lambda r: abs(r.filled_area - exact_size)
            )[0]
            y_dist = most_similar_bubble.centroid[0] - slice.phan_center.y
            x_dist = most_similar_bubble.centroid[1] - slice.phan_center.x
            phan_roll = np.arctan2(y_dist, x_dist)
            corrected_roll = (
                np.rad2deg(phan_roll) + 135
            )  # bubble is at top-left. perfect placement is -135
            return corrected_roll
        except:
            raise RuntimeError(
                "Could not determine the roll of the phantom. Ensure the 20mm top-left circle is visible on Slice 1"
            )

    def analyze(self) -> None:
        """Analyze the ACR CT phantom"""
        self.slice1 = MRSlice1Module(self, offset=0)
        self.geometric_distortion = GeometricDistortionModule(
            self, offset=MR_GEOMETRIC_DISTORTION_MODULE_OFFSET_MM
        )
        self.uniformity_module = MRUniformityModule(
            self, offset=MR_UNIFORMITY_MODULE_OFFSET_MM
        )
        self.slice11 = MRSlice11PositionModule(self, offset=MR_SLICE11_MODULE_OFFSET_MM)

    def plot_analyzed_image(self, show: bool = True, **plt_kwargs) -> plt.Figure:
        """Plot the analyzed image

        Parameters
        ----------
        show
            Whether to show the image.
        plt_kwargs
            Keywords to pass to matplotlib for figure customization.
        """
        # set up grid and axes
        fig = plt.figure(**plt_kwargs)
        grid_size = (2, 3)
        slice1_ax = plt.subplot2grid(grid_size, (0, 0))
        self.slice1.plot(slice1_ax)
        geom_ax = plt.subplot2grid(grid_size, (0, 1))
        self.geometric_distortion.plot(geom_ax)
        unif_ax = plt.subplot2grid(grid_size, (0, 2))
        self.uniformity_module.plot(unif_ax)
        position_ax = plt.subplot2grid(grid_size, (1, 0))
        self.slice11.plot(position_ax)

        spatial_res_graph = plt.subplot2grid(grid_size, (1, 1))
        self.slice1.row_mtf.plot(spatial_res_graph, label="Row-wise rMTF")
        self.slice1.col_mtf.plot(spatial_res_graph, label="Column-wise rMTF")
        spatial_res_graph.legend()

        # finish up
        plt.tight_layout()
        if show:
            plt.show()
        return fig

    def plot_images(self, show: bool = True, **plt_kwargs) -> Dict[str, plt.Figure]:
        """Plot all the individual images separately

        Parameters
        ----------
        show
            Whether to show the images.
        plt_kwargs
            Keywords to pass to matplotlib for figure customization.
        """
        figs = {}
        # plot the images
        modules = {
            "geometric": self.geometric_distortion,
            "slice 1": self.slice1,
            "signal uniformity": self.uniformity_module,
            "slice 11": self.slice11,
        }
        for key, module in modules.items():
            fig, ax = plt.subplots(**plt_kwargs)
            module.plot(ax)
            figs[key] = fig
        # plot rMTF
        fig, ax = plt.subplots(**plt_kwargs)
        self.slice1.row_mtf.plot(ax, label="Row-wise rMTF")
        self.slice1.col_mtf.plot(ax, label="Column-wise rMTF")
        ax.legend()
        figs["rMTF"] = fig

        if show:
            plt.show()
        return figs

    def save_images(
        self,
        directory: Optional[Union[Path, str]] = None,
        to_stream: bool = False,
        **plt_kwargs,
    ) -> List[Union[Path, BytesIO]]:
        """Save separate images to disk or stream.

        Parameters
        ----------
        directory
            The directory to write the images to. If None, will use current working directory
        to_stream
            Whether to write to stream or disk. If True, will return streams. Directory is ignored in that scenario.
        plt_kwargs
            Keywords to pass to matplotlib for figure customization.
        """
        figs = self.plot_images(show=False, **plt_kwargs)
        paths = []
        for name, fig in figs.items():
            if to_stream:
                path = io.BytesIO()
            else:
                destination = Path(directory) or Path.cwd()
                path = (destination / name).with_suffix(".png").absolute()
            fig.savefig(path)
            paths.append(path)
        return paths

    def publish_pdf(
        self,
        filename: Union[str, Path],
        notes: Optional[str] = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ) -> None:
        """Publish (print) a PDF containing the analysis and quantitative results.

        Parameters
        ----------
        filename : (str, file-like object}
            The file to write the results to.
        notes : str, list of strings
            Text; if str, prints single line.
            If list of strings, each list item is printed on its own line.
        open_file : bool
            Whether to open the file using the default program after creation.
        metadata : dict
            Extra data to be passed and shown in the PDF. The key and value will be shown with a colon.
            E.g. passing {'Author': 'James', 'Unit': 'TrueBeam'} would result in text in the PDF like:
            --------------
            Author: James
            Unit: TrueBeam
            --------------
        """
        analysis_title = f"{self._model} Analysis"
        analysis_images = self.save_images(to_stream=True)

        canvas = pdf.PylinacCanvas(
            filename, page_title=analysis_title, metadata=metadata
        )
        if notes is not None:
            canvas.add_text(text="Notes:", location=(1, 4.5), font_size=14)
            canvas.add_text(text=notes, location=(1, 4))

        shortened_texts = [
            textwrap.wrap(r, width=110) for r in self.results(as_str=False)
        ]
        idx = 0
        for items in enumerate(shortened_texts):
            for text in items:
                canvas.add_text(text=text, location=(1.5, 25 - idx * 0.5))
                idx += 1
        for page, img in enumerate(analysis_images):
            canvas.add_new_page()
            canvas.add_image(img, location=(1, 5), dimensions=(18, 18))
        canvas.finish()

        if open_file:
            webbrowser.open(filename)

    def results(self, as_str: bool = True) -> Union[str, Tuple]:
        """Return the results of the analysis as a string. Use with print()."""
        string = (
            f" - {self._model} Results - ",
            f"Geometric Distortions: {self.geometric_distortion.distances()}",
            f"Slice Thickness: {self.slice1.slice_thickness:2.2f}mm",
            f"Slice 1 S/I Position shift: {self.slice1.slice_shift_mm:2.2f}mm",
            f"Slice 11 S/I Position shift: {self.slice11.slice_shift_mm:2.2f}mm",
            f"Uniformity PIU: {self.uniformity_module.percent_image_uniformity:2.2f}",
            f"Percent-signal ghosting: {self.uniformity_module.psg:2.2f}%",
            f'Uniformity Center ROI standard deviation: {self.uniformity_module.rois["Center"].std:2.2f}',
            f"Row-wise MTF 50% (lp/mm): {self.slice1.row_mtf.relative_resolution(50):2.2f}",
            f"Column-wise MTF 50% (lp/mm): {self.slice1.col_mtf.relative_resolution(50):2.2f}",
        )
        if as_str:
            return "\n".join(string)
        else:
            return string

    def results_data(self, as_dict: bool = False) -> Union[ACRMRIResult, dict]:
        data = ACRMRIResult(
            phantom_model=self._model,
            phantom_roll_deg=self.catphan_roll,
            origin_slice=self.origin_slice,
            num_images=self.num_images,
            slice1=MRSlice1ModuleOutput(
                offset=0,
                roi_settings=self.slice1.roi_settings,
                rois=self.slice1.rois,
                bar_difference_mm=self.slice1.bar_difference_mm,
                slice_shift_mm=self.slice1.slice_shift_mm,
                measured_slice_thickness_mm=self.slice1.measured_slice_thickness_mm,
                row_mtf_50=self.slice1.row_mtf.relative_resolution(50),
                col_mtf_50=self.slice1.col_mtf.relative_resolution(50),
            ),
            slice11=MRSlice11ModuleOutput(
                offset=MR_SLICE11_MODULE_OFFSET_MM,
                bar_difference_mm=self.slice11.bar_difference_mm,
                slice_shift_mm=self.slice11.slice_shift_mm,
                rois=self.slice11.rois,
                roi_settings=self.slice11.roi_settings,
            ),
            geometric_distortion_module=MRGeometricDistortionModuleOutput(
                offset=MR_GEOMETRIC_DISTORTION_MODULE_OFFSET_MM,
                profiles=self.geometric_distortion.profiles,
                distances=self.geometric_distortion.distances(),
            ),
            uniformity_module=MRUniformityModuleOutput(
                offset=0,
                roi_settings=self.uniformity_module.roi_settings,
                rois=self.uniformity_module.rois,
                ghost_roi_settings=self.uniformity_module.ghost_roi_settings,
                ghose_rois=self.uniformity_module.ghost_rois,
                psg=self.uniformity_module.psg,
                ghosting_ratio=self.uniformity_module.ghosting_ratio,
                piu=self.uniformity_module.percent_image_uniformity,
                piu_passed=self.uniformity_module.piu_passed,
            ),
        )
        if as_dict:
            return dataclasses.asdict(data)
        return data


def find_nearest_idx(array: np.ndarray, value: float):
    return (np.abs(array - value)).argmin()
